#!/usr/bin/env -S PYTHONPATH=.. pytest-3 -v

"""
Calling tests.

This test script can be invoked in three ways:

- By itself, in your PC's current Python environment, without relying on a
  running CRS board:

      ./test_calls.py

- By itself, in your PC's current Python environment, with a CRS board running
  alongside:

      ./test_calls.py --serial=0024

- As part of a complete regression test environment (exercising all test
  scripts and in a variety of different Python environments), without relying
  on a running CRS board:

      ~/rfmux$ ./test.sh
"""

import rfmux
import pytest


@pytest.mark.asyncio
async def test_macro():
    @rfmux.macro(rfmux.CRS, register=True)
    async def macro_that_returns_serial_number(d):
        return d.serial

    s = rfmux.load_session(
        """
        !HardwareMap
        - !CRS { serial: "0024" }
        """
    )
    ds = s.query(rfmux.CRS)
    assert set(await ds.macro_that_returns_serial_number()) == {"0024"}

    d = s.query(rfmux.CRS).one()
    assert d.serial == await d.macro_that_returns_serial_number()


@pytest.mark.asyncio
async def test_algorithm():
    @rfmux.algorithm(rfmux.CRS)
    async def algorithm_that_returns_serial_number(d):
        return d.serial

    s = rfmux.load_session(
        """
        !HardwareMap
        - !CRS { serial: "0024" }
        """
    )
    ds = s.query(rfmux.CRS)
    assert set(await ds.algorithm_that_returns_serial_number()) == {"0024"}


@pytest.mark.asyncio
async def test_simple_live_board_interaction(crs):
    await crs.get_frequency(crs.UNITS.HZ, channel=1, module=1)


@pytest.mark.asyncio
async def test_live_board_interaction_with_orm(live_session):
    ds = live_session.query(rfmux.CRS)
    await ds.resolve()

    d = live_session.query(rfmux.CRS).one()
    f = await d.get_frequency(d.UNITS.HZ, channel=1, module=1)
    fs = await ds.get_frequency(d.UNITS.HZ, channel=1, module=1)

    assert f == fs[0]
    assert await d.module[1].channel[1].get_frequency(d.UNITS.HZ) == f


@pytest.mark.asyncio
async def test_macro_with_arg_filler():
    @rfmux.macro(rfmux.CRS, register=True)
    async def channel_macro(self, channel, module):
        return module == 3 and channel == 4

    s = rfmux.load_session('!HardwareMap [!CRS { serial: "0024" }]')
    d = s.query(rfmux.CRS).one()

    assert await d.module[3].channel[4].channel_macro()


if __name__ == "__main__":
    pytest.main([__file__])



"""
Session Browser Panel for Periscope Application
================================================

This module provides a dockable file browser panel for viewing and
loading session files. It displays the contents of the current session
folder using a QTreeView with QFileSystemModel.

Features:
- Displays session files in a tree view
- Double-click to load files into new panels
- Refresh button to update file list
- Open folder button to show in system file browser
- Filter by file type
"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Optional

from PyQt6 import QtCore, QtWidgets, QtGui

if TYPE_CHECKING:
    from .session_manager import SessionManager


class SessionBrowserPanel(QtWidgets.QWidget):
    """
    A dockable panel displaying session files with a tree view.
    
    This panel provides a file browser interface for viewing the contents
    of the current session folder. Users can double-click files to load
    them into new analysis panels.
    
    Signals:
        file_load_requested: Emitted when user double-clicks a file (file_path: str)
    
    Attributes:
        session_manager: Reference to the SessionManager instance
    """
    
    # Signal emitted when user wants to load a file
    file_load_requested = QtCore.pyqtSignal(str)  # file_path
    
    def __init__(self, session_manager: SessionManager, parent: Optional[QtWidgets.QWidget] = None):
        """
        Initialize the SessionBrowserPanel.
        
        Args:
            session_manager: The SessionManager instance to monitor
            parent: Optional parent widget
        """
        super().__init__(parent)
        
        self.session_manager = session_manager
        self._dark_mode = False
        
        self._setup_ui()
        self._connect_signals()
        self._update_session_display()
    
    def _setup_ui(self):
        """Create and configure the panel UI."""
        layout = QtWidgets.QVBoxLayout(self)
        layout.setContentsMargins(5, 5, 5, 5)
        layout.setSpacing(5)
        
        # ─────────────────────────────────────────────────────────────
        # Header section with session info
        # ─────────────────────────────────────────────────────────────
        
        header_group = QtWidgets.QGroupBox("Session")
        header_layout = QtWidgets.QVBoxLayout(header_group)
        header_layout.setContentsMargins(5, 5, 5, 5)
        
        # Session status label
        self.status_label = QtWidgets.QLabel("No active session")
        self.status_label.setWordWrap(True)
        self.status_label.setStyleSheet("font-weight: bold;")
        header_layout.addWidget(self.status_label)
        
        # Session path label (truncated if long)
        self.path_label = QtWidgets.QLabel("")
        self.path_label.setWordWrap(True)
        self.path_label.setStyleSheet("font-size: 10px; color: gray;")
        self.path_label.setTextInteractionFlags(
            QtCore.Qt.TextInteractionFlag.TextSelectableByMouse
        )
        header_layout.addWidget(self.path_label)
        
        # File count label
        self.count_label = QtWidgets.QLabel("")
        self.count_label.setStyleSheet("font-size: 10px;")
        header_layout.addWidget(self.count_label)
        
        layout.addWidget(header_group)
        
        # ─────────────────────────────────────────────────────────────
        # Filter section
        # ─────────────────────────────────────────────────────────────
        
        filter_layout = QtWidgets.QHBoxLayout()
        filter_layout.setContentsMargins(0, 0, 0, 0)
        
        filter_label = QtWidgets.QLabel("Filter:")
        filter_layout.addWidget(filter_label)
        
        self.filter_combo = QtWidgets.QComboBox()
        self.filter_combo.addItem("All Files", "all")
        self.filter_combo.addItem("Notebooks (.ipynb)", "ipynb")
        self.filter_combo.addItem("Network Analysis", "netanal")
        self.filter_combo.addItem("Multisweep", "multisweep")
        self.filter_combo.addItem("Bias KIDs", "bias")
        self.filter_combo.addItem("Noise Spectrum", "noise")
        self.filter_combo.addItem("Screenshots", "screenshot")
        self.filter_combo.currentIndexChanged.connect(self._on_filter_changed)
        filter_layout.addWidget(self.filter_combo, 1)
        
        layout.addLayout(filter_layout)
        
        # ─────────────────────────────────────────────────────────────
        # File tree view
        # ─────────────────────────────────────────────────────────────
        
        self.tree_view = QtWidgets.QTreeView()
        self.tree_view.setAlternatingRowColors(True)
        self.tree_view.setRootIsDecorated(False)
        self.tree_view.setSortingEnabled(True)
        self.tree_view.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.SingleSelection)
        self.tree_view.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
        
        # Set up file system model (QFileSystemModel is in QtGui)
        self.file_model = QtGui.QFileSystemModel()
        self.file_model.setNameFilters(["*.pkl", "*.ipynb", ".png"])
        self.file_model.setNameFilterDisables(False)
        self.file_model.setFilter(
            QtCore.QDir.Filter.Files | QtCore.QDir.Filter.NoDotAndDotDot
        )
        
        # IMPORTANT: Disable automatic file watching to prevent race conditions
        # with Jupyter Lab's file watcher running inside QWebEngineView.
        # Both watchers reacting to the same file system events can cause segfaults
        # in Qt's accessibility system (QAccessible::registerAccessibleInterface).
        # We use a gentle polling timer instead for automatic updates.
        self.file_model.setOption(
            QtGui.QFileSystemModel.Option.DontWatchForChanges, True
        )
        
        self.tree_view.setModel(self.file_model)
        
        # Hide unnecessary columns (keep Name, Size, Date Modified)
        self.tree_view.hideColumn(2)  # Hide Type column
        
        # Set column widths
        self.tree_view.setColumnWidth(0, 200)  # Name
        self.tree_view.setColumnWidth(1, 60)   # Size
        self.tree_view.setColumnWidth(3, 120)  # Date Modified
        
        # Sort by Date Modified descending (newest first)
        self.tree_view.sortByColumn(3, QtCore.Qt.SortOrder.DescendingOrder)
        
        # Connect double-click
        self.tree_view.doubleClicked.connect(self._on_file_double_clicked)
        self.tree_view.customContextMenuRequested.connect(self._show_context_menu)
        
        layout.addWidget(self.tree_view, 1)  # Stretch factor 1
        
        # ─────────────────────────────────────────────────────────────
        # Button row
        # ─────────────────────────────────────────────────────────────
        
        button_layout = QtWidgets.QHBoxLayout()
        button_layout.setContentsMargins(0, 0, 0, 0)
        
        self.refresh_btn = QtWidgets.QPushButton("Refresh")
        self.refresh_btn.setToolTip("Refresh the file list")
        self.refresh_btn.clicked.connect(self.refresh)
        button_layout.addWidget(self.refresh_btn)
        
        self.open_folder_btn = QtWidgets.QPushButton("Open Folder")
        self.open_folder_btn.setToolTip("Open session folder in file browser")
        self.open_folder_btn.clicked.connect(self._on_open_folder_clicked)
        button_layout.addWidget(self.open_folder_btn)
        
        layout.addLayout(button_layout)
        
        # ─────────────────────────────────────────────────────────────
        # Placeholder for when no session is active
        # ─────────────────────────────────────────────────────────────
        
        self.placeholder_label = QtWidgets.QLabel(
            "No session active.\n\n"
            "Use Session → Start New Session\n"
            "to begin auto-exporting data."
        )
        self.placeholder_label.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.placeholder_label.setStyleSheet("color: gray; padding: 20px;")
        self.placeholder_label.setWordWrap(True)
        
        # Initially show placeholder, hide tree
        self.tree_view.setVisible(False)
        self.filter_combo.setEnabled(False)
        self.refresh_btn.setEnabled(False)
        self.open_folder_btn.setEnabled(False)
        
        layout.addWidget(self.placeholder_label)
        
        # ─────────────────────────────────────────────────────────────
        # Polling timer for gentle automatic refresh
        # ─────────────────────────────────────────────────────────────
        # Since we disabled QFileSystemModel's file watcher (to prevent race
        # conditions with Jupyter Lab), we use a gentle polling timer instead.
        # This provides automatic updates without the segfault-inducing races.
        self._poll_timer = QtCore.QTimer(self)
        self._poll_timer.timeout.connect(self._poll_refresh)
        self._poll_timer.setInterval(5000)  # 5 second interval - generous polling
    
    def _connect_signals(self):
        """Connect signals from the session manager."""
        self.session_manager.session_started.connect(self._on_session_started)
        self.session_manager.session_ended.connect(self._on_session_ended)
        self.session_manager.session_updated.connect(self.refresh)
        self.session_manager.file_exported.connect(self._on_file_exported)
    
    # ─────────────────────────────────────────────────────────────────
    # Public Methods
    # ─────────────────────────────────────────────────────────────────
    
    def refresh(self):
        """Refresh the file list display."""
        if not self.session_manager.is_active:
            return
        
        root_path = str(self.session_manager.session_path)
        
        # With DontWatchForChanges enabled, QFileSystemModel caches directory contents.
        # To force a true refresh, we need to reset the root path by setting it
        # to an empty string first, then back to the actual path.
        self.file_model.setRootPath("")
        QtCore.QCoreApplication.processEvents()
        self.file_model.setRootPath(root_path)
        self.tree_view.setRootIndex(self.file_model.index(root_path))
        
        # Apply current filter
        self._apply_filter()
        
        # Update count label
        self._update_file_count()
    
    def apply_theme(self, dark_mode: bool):
        """
        Apply the dark/light theme to the panel.
        
        Args:
            dark_mode: True for dark theme, False for light theme
        """
        self._dark_mode = dark_mode
        
        if dark_mode:
            self.path_label.setStyleSheet("font-size: 10px; color: #aaaaaa;")
            self.placeholder_label.setStyleSheet("color: #888888; padding: 20px;")
        else:
            self.path_label.setStyleSheet("font-size: 10px; color: gray;")
            self.placeholder_label.setStyleSheet("color: gray; padding: 20px;")
    
    # ─────────────────────────────────────────────────────────────────
    # Session State Handlers
    # ─────────────────────────────────────────────────────────────────
    
    @QtCore.pyqtSlot(str)
    def _on_session_started(self, session_path: str):
        """Handle session start."""
        self._update_session_display()
        
        # Set the file model root to the session path
        self.file_model.setRootPath(session_path)
        self.tree_view.setRootIndex(self.file_model.index(session_path))
        
        # Show tree, hide placeholder
        self.tree_view.setVisible(True)
        self.placeholder_label.setVisible(False)
        
        # Enable controls
        self.filter_combo.setEnabled(True)
        self.refresh_btn.setEnabled(True)
        self.open_folder_btn.setEnabled(True)
        
        self._update_file_count()
        
        # Start the polling timer for automatic refresh
        self._poll_timer.start()
    
    @QtCore.pyqtSlot()
    def _on_session_ended(self):
        """Handle session end."""
        # Stop the polling timer
        self._poll_timer.stop()
        
        self._update_session_display()
        
        # Clear the tree
        self.file_model.setRootPath("")
        
        # Hide tree, show placeholder
        self.tree_view.setVisible(False)
        self.placeholder_label.setVisible(True)
        
        # Disable controls
        self.filter_combo.setEnabled(False)
        self.refresh_btn.setEnabled(False)
        self.open_folder_btn.setEnabled(False)
    
    @QtCore.pyqtSlot(str, str)
    def _on_file_exported(self, file_path: str, data_type: str):
        """Handle new file export - refresh and highlight."""
        self.refresh()
        
        # Try to select the newly exported file
        index = self.file_model.index(file_path)
        if index.isValid():
            self.tree_view.setCurrentIndex(index)
            self.tree_view.scrollTo(index)
    
    def _poll_refresh(self):
        """
        Gentle polling refresh - called periodically by the timer.
        
        Only refreshes when session is active and panel is visible.
        This replaces QFileSystemModel's internal file watcher to avoid
        race conditions with Jupyter Lab's file watcher.
        """
        if self.session_manager.is_active and self.isVisible():
            self.refresh()
    
    def _update_session_display(self):
        """Update the session status labels."""
        if self.session_manager.is_active:
            self.status_label.setText(f"📁 {self.session_manager.session_name}")
            self.status_label.setStyleSheet("font-weight: bold; color: green;")
            
            # Show parent folder + session name for clarity
            full_path = Path(self.session_manager.session_path)
            display_path = str(Path(full_path.parent.name) / full_path.name)
            self.path_label.setText(display_path)
            self.path_label.setToolTip(str(self.session_manager.session_path))
        else:
            self.status_label.setText("No active session")
            self.status_label.setStyleSheet("font-weight: bold;")
            self.path_label.setText("")
            self.path_label.setToolTip("")
            self.count_label.setText("")
    
    def _update_file_count(self):
        """Update the file count label."""
        if not self.session_manager.is_active:
            self.count_label.setText("")
            return
        
        summary = self.session_manager.get_session_summary()
        total = summary.get('total_files', 0)
        
        if total == 0:
            self.count_label.setText("No files yet")
        elif total == 1:
            self.count_label.setText("1 file")
        else:
            self.count_label.setText(f"{total} files")
    
    # ─────────────────────────────────────────────────────────────────
    # Filter Handling
    # ─────────────────────────────────────────────────────────────────
    
    def _on_filter_changed(self, index: int):
        """Handle filter combo box changes."""
        self._apply_filter()
    
    def _apply_filter(self):
        """Apply the current filter to the file model."""
        filter_type = self.filter_combo.currentData()
        
        if filter_type == "all":
            self.file_model.setNameFilters(["*.pkl", "*.ipynb", "*.png"])
        elif filter_type == "ipynb":
            self.file_model.setNameFilters(["*.ipynb"])
        elif filter_type == "screenshot":
            self.file_model.setNameFilters(["*.png"])
        else:
            # Filter by data type in filename (new format: type_identifier_HHMMSS.pkl)
            # Also support old format for backwards compatibility
            self.file_model.setNameFilters([f"{filter_type}_*.pkl", f"*_{filter_type}_*.pkl"])
    
    # ─────────────────────────────────────────────────────────────────
    # File Operations
    # ─────────────────────────────────────────────────────────────────
    
    def _on_file_double_clicked(self, index: QtCore.QModelIndex):
        """Handle double-click on a file."""
        if not index.isValid():
            return
        
        file_path = self.file_model.filePath(index)
        
        # Check it's a file (not a directory)
        if not Path(file_path).is_file():
            return
        
        # Emit signal to request file load
        self.file_load_requested.emit(file_path)
    
    def _on_open_folder_clicked(self):
        """Handle Open Folder button click."""
        self.session_manager.open_session_folder()
    
    def _show_context_menu(self, position: QtCore.QPoint):
        """Show context menu for file operations."""
        index = self.tree_view.indexAt(position)
        if not index.isValid():
            return
        
        file_path = self.file_model.filePath(index)
        if not Path(file_path).is_file():
            return
        
        menu = QtWidgets.QMenu(self)
        
        # Load action
        load_action = menu.addAction("Load File")
        load_action.triggered.connect(lambda: self.file_load_requested.emit(file_path))
        
        menu.addSeparator()
        
        # Copy path action
        copy_path_action = menu.addAction("Copy Path")
        copy_path_action.triggered.connect(
            lambda: QtWidgets.QApplication.clipboard().setText(file_path)
        )
        
        # Show in folder action
        show_in_folder_action = menu.addAction("Show in Folder")
        show_in_folder_action.triggered.connect(
            lambda: self._show_file_in_folder(file_path)
        )
        
        menu.addSeparator()

        rename_action = menu.addAction("Rename File")
        rename_action.triggered.connect(lambda: self._rename_file(file_path))
        
        # Delete action
        delete_action = menu.addAction("Delete File")
        delete_action.triggered.connect(lambda: self._delete_file(file_path))
        
        menu.exec(self.tree_view.viewport().mapToGlobal(position))
    
    def _show_file_in_folder(self, file_path: str):
        """Open the containing folder and select the file."""
        import subprocess
        import sys
        
        try:
            if sys.platform == 'darwin':  # macOS
                subprocess.run(['open', '-R', file_path], check=True)
            elif sys.platform == 'win32':  # Windows
                subprocess.run(['explorer', '/select,', file_path], check=True)
            else:  # Linux
                # xdg-open opens the folder, not selecting the file
                folder = str(Path(file_path).parent)
                subprocess.run(['xdg-open', folder], check=True)
        except Exception as e:
            print(f"[SessionBrowser] Error showing file in folder: {e}")
    
    
    def _rename_file(self, file_path: str):
        """Rename a file with user input and validation."""
        path = Path(file_path)
        old_name = path.name

        new_name, ok = QtWidgets.QInputDialog.getText(
            self,
            "Rename File",
            "Enter new file name:",
            QtWidgets.QLineEdit.EchoMode.Normal,
            old_name
        )

        if not ok or not new_name.strip():
            return

        new_name = new_name.strip()
        new_path = path.with_name(new_name)

        if new_path == path:
            return

        if new_path.exists():
            QtWidgets.QMessageBox.warning(
                self,
                "Rename Error",
                "A file with this name already exists."
            )
            return

        try:
            path.rename(new_path)
            self.refresh()
            print(f"[SessionBrowser] Renamed: {old_name} → {new_name}")
        except Exception as e:
            QtWidgets.QMessageBox.critical(
                self,
                "Rename Error",
                f"Could not rename file:\n{str(e)}"
            )

    
    def _delete_file(self, file_path: str):
        """Delete a file with confirmation."""
        filename = Path(file_path).name
        
        reply = QtWidgets.QMessageBox.question(
            self,
            "Delete File",
            f"Are you sure you want to delete:\n{filename}?",
            QtWidgets.QMessageBox.StandardButton.Yes | QtWidgets.QMessageBox.StandardButton.No,
            QtWidgets.QMessageBox.StandardButton.No
        )
        
        if reply == QtWidgets.QMessageBox.StandardButton.Yes:
            try:
                Path(file_path).unlink()
                self.refresh()
                print(f"[SessionBrowser] Deleted: {filename}")
            except Exception as e:
                QtWidgets.QMessageBox.critical(
                    self,
                    "Delete Error",
                    f"Could not delete file:\n{str(e)}"
                )



"""
Periscope Settings Module
=========================

Centralized QSettings wrapper for remembering user preferences across sessions.

Uses Qt's QSettings which persists settings to:
- Windows: Registry (HKEY_CURRENT_USER\\Software\\rfmux\\periscope)
- macOS: ~/Library/Preferences/com.rfmux.periscope.plist
- Linux: ~/.config/rfmux/periscope.conf

Usage:
    from . import settings
    
    # Get/set last session directory
    last_dir = settings.get_last_session_directory()
    settings.set_last_session_directory("/path/to/sessions")
"""

from PyQt6.QtCore import QSettings
from pathlib import Path
from typing import Optional

# Organization and application name for QSettings
ORGANIZATION = "rfmux"
APPLICATION = "periscope"

# Settings keys
KEY_CONNECTION_MODE = "connection/last_mode"
KEY_CRS_SERIAL = "connection/last_crs_serial"
KEY_MODULE = "connection/last_module"
KEY_SESSION_DIRECTORY = "session/last_base_directory"
KEY_USER_LIBRARY_PATH = "notebook/user_library_path"
KEY_CUSTOM_MATERIALS = "materials/custom_materials"

# Default values
DEFAULT_CONNECTION_MODE = "hardware"
DEFAULT_MODULE = 1


def _get_settings() -> QSettings:
    """Get a QSettings instance with our org/app names."""
    return QSettings(ORGANIZATION, APPLICATION)


# ─────────────────────────────────────────────────────────────────
# Connection Settings
# ─────────────────────────────────────────────────────────────────

def get_last_connection_mode() -> str:
    """
    Get the last used connection mode.
    
    Returns:
        str: "hardware", "mock", or "offline"
    """
    settings = _get_settings()
    return settings.value(KEY_CONNECTION_MODE, DEFAULT_CONNECTION_MODE)


def set_last_connection_mode(mode: str) -> None:
    """
    Save the connection mode.
    
    Args:
        mode: "hardware", "mock", or "offline"
    """
    if mode not in ("hardware", "mock", "offline"):
        raise ValueError(f"Invalid connection mode: {mode}")
    settings = _get_settings()
    settings.setValue(KEY_CONNECTION_MODE, mode)


def get_last_crs_serial() -> str:
    """
    Get the last used CRS serial number.
    
    Returns:
        str: CRS serial number (e.g., "0042") or empty string
    """
    settings = _get_settings()
    return settings.value(KEY_CRS_SERIAL, "")


def set_last_crs_serial(serial: str) -> None:
    """
    Save the CRS serial number.
    
    Args:
        serial: CRS serial number (e.g., "0042")
    """
    settings = _get_settings()
    settings.setValue(KEY_CRS_SERIAL, serial)


def get_last_module() -> int:
    """
    Get the last used module number.
    
    Returns:
        int: Module number (1-8), defaults to 1
    """
    settings = _get_settings()
    value = settings.value(KEY_MODULE, DEFAULT_MODULE)
    # QSettings may return string on some platforms
    return int(value) if value else DEFAULT_MODULE


def set_last_module(module: int) -> None:
    """
    Save the module number.
    
    Args:
        module: Module number (1-8)
    """
    if not 1 <= module <= 8:
        raise ValueError(f"Module must be 1-8, got: {module}")
    settings = _get_settings()
    settings.setValue(KEY_MODULE, module)


# ─────────────────────────────────────────────────────────────────
# Session Settings
# ─────────────────────────────────────────────────────────────────

def get_last_session_directory() -> str:
    """
    Get the last directory used for session creation/loading.
    
    Returns:
        str: Directory path or empty string
    """
    settings = _get_settings()
    return settings.value(KEY_SESSION_DIRECTORY, "")


def set_last_session_directory(path: str) -> None:
    """
    Save the session directory.
    
    Args:
        path: Directory path
    """
    settings = _get_settings()
    settings.setValue(KEY_SESSION_DIRECTORY, str(path))


# ─────────────────────────────────────────────────────────────────
# Notebook Settings
# ─────────────────────────────────────────────────────────────────

def get_user_library_path() -> str:
    """
    Get the user notebook library path (used for symlinks in sessions).
    
    Returns:
        str: Library path or empty string
    """
    settings = _get_settings()
    return settings.value(KEY_USER_LIBRARY_PATH, "")


def set_user_library_path(path: str) -> None:
    """
    Save the user notebook library path.
    
    Args:
        path: Library directory path
    """
    settings = _get_settings()
    settings.setValue(KEY_USER_LIBRARY_PATH, str(path))


# ─────────────────────────────────────────────────────────────────
# Utility Functions
# ─────────────────────────────────────────────────────────────────

def clear_all() -> None:
    """Clear all saved settings (useful for testing/reset)."""
    settings = _get_settings()
    settings.clear()


def get_settings_path() -> Optional[Path]:
    """
    Get the path to the settings file (for debugging).
    
    Returns:
        Path to settings file, or None if using registry (Windows)
    """
    settings = _get_settings()
    filename = settings.fileName()
    return Path(filename) if filename else None


# ─────────────────────────────────────────────────────────────────
# Custom Materials Settings
# ─────────────────────────────────────────────────────────────────

def get_custom_materials() -> dict:
    """
    Get all user-defined custom materials.
    
    Returns:
        dict: Material name -> properties dict
              e.g., {'Nb': {'Tc': 9.2, 'N0': 3.5e10, 'tau0': 200e-9, 'sigmaN': 5e7}}
    """
    import json
    settings = _get_settings()
    json_str = settings.value(KEY_CUSTOM_MATERIALS, "{}")
    try:
        return json.loads(json_str)
    except (json.JSONDecodeError, TypeError):
        return {}


def save_custom_material(name: str, tc: float, n0: float, tau0: float, Rs: Optional[float] = None, thickness_ref_nm: Optional[float] = None, sigmaN: Optional[float] = None) -> None:
    """
    Add or update a custom material.
    
    Args:
        name: Material name/symbol (e.g., "Nb", "TiN")
        tc: Critical temperature [K]
        n0: Density of states at Fermi level [µm⁻³eV⁻¹]
        tau0: Quasiparticle recombination time [s]
        Rs: Sheet resistance [Ω/□] - preferred input method
        thickness_ref_nm: Reference film thickness [nm] where Rs was measured
        sigmaN: Normal state conductivity [S/m] - calculated from Rs if not provided
        
    Note:
        Either (Rs AND thickness_ref_nm) OR sigmaN must be provided.
        If Rs is provided, sigmaN is calculated as: σN = 1/(Rs × thickness)
    """
    import json
    materials = get_custom_materials()
    
    # Calculate sigmaN from Rs if provided
    if Rs is not None and thickness_ref_nm is not None:
        thickness_m = thickness_ref_nm * 1e-9  # nm to m
        sigmaN = 1.0 / (Rs * thickness_m)
        
        mat_data = {
            'Tc': float(tc),
            'N0': float(n0),
            'tau0': float(tau0),
            'Rs': float(Rs),
            'thickness_ref': float(thickness_ref_nm),
            'sigmaN': float(sigmaN)
        }
    elif sigmaN is not None:
        # Legacy: direct sigmaN input (backward compatibility)
        mat_data = {
            'Tc': float(tc),
            'N0': float(n0),
            'tau0': float(tau0),
            'sigmaN': float(sigmaN)
        }
    else:
        # Default Rs and thickness for backward compatibility
        Rs_default = 4.0
        thickness_default = 20.0
        sigmaN_default = 1.0 / (Rs_default * thickness_default * 1e-9)
        mat_data = {
            'Tc': float(tc),
            'N0': float(n0),
            'tau0': float(tau0),
            'Rs': Rs_default,
            'thickness_ref': thickness_default,
            'sigmaN': sigmaN_default
        }
    
    materials[name] = mat_data
    
    settings = _get_settings()
    settings.setValue(KEY_CUSTOM_MATERIALS, json.dumps(materials))


def delete_custom_material(name: str) -> None:
    """
    Remove a custom material.
    
    Args:
        name: Material name to remove
    """
    import json
    materials = get_custom_materials()
    if name in materials:
        del materials[name]
        settings = _get_settings()
        settings.setValue(KEY_CUSTOM_MATERIALS, json.dumps(materials))


def get_material_properties(name: str) -> dict:
    """
    Get properties for a material (built-in or custom).
    
    Args:
        name: Material name (e.g., "Al", "Nb")
        
    Returns:
        dict: {
            'Tc': float,      # Critical temperature [K]
            'N0': float,      # Density of states [µm⁻³eV⁻¹] - NOTE: eV units, needs conversion!
            'tau0': float,    # Quasiparticle recombination time [s]
            'sigmaN': float   # Normal state conductivity [S/m]
        }
        
    Note:
        N0 is stored in µm⁻³eV⁻¹ for human readability. When passing to
        MR_complex_resonator, convert to µm⁻³J⁻¹:
        
            N0_internal = N0_eV / 1.602e-19
        
    Raises:
        ValueError: If material not found
    """
    # Built-in materials
    if name == "Al":
        return {
            'Tc': 1.2,
            'N0': 1.72e10,  # µm⁻³eV⁻¹
            'tau0': 438e-9,  # s
            'Rs': 4.0,  # Ω/□ (typical for thin Al)
            'thickness_ref': 20.0,  # nm (reference thickness)
            'sigmaN': 1./(4*20e-9)  # S/m (calculated from Rs)
        }
    
    # Custom materials
    materials = get_custom_materials()
    if name in materials:
        mat = materials[name].copy()
        # Ensure sigmaN has default if not specified
        if 'sigmaN' not in mat:
            mat['sigmaN'] = 1./(4*20e-9)
        return mat
    
    raise ValueError(f"Material '{name}' not found in built-in or custom materials")



__all__ = [
    "core",
    "algorithms",
    "tools",
    "streamer",
    "Crate",
    "CRS",
    "ReadoutModule",
    "ReadoutChannel",
    "macro",
    "algorithm",
    "HardwareMap",
    "load_dirfile",
    "load_session",
    "get_session",
    "set_session",
    "load_from_database",
    "TuberRemoteError",
]

# Version from setuptools-scm
try:
    from ._version import __version__
except ImportError:
    __version__ = "unknown"

# Check version numbers
import sys

if sys.version_info < (3, 10):
    raise RuntimeError("Python >= 3.10 is required.")

import importlib.metadata  # requires Python 3.8+
from importlib.metadata import PackageNotFoundError
from packaging.version import parse

tuples = (("sqlalchemy", "2.0.0"), ("IPython", "8.0.0"))

for package, min_version in tuples:
    try:
        installed_version = importlib.metadata.version(package)
    except PackageNotFoundError as e:
        raise PackageNotFoundError(f"Package {package} is not installed!")

    if parse(installed_version) < parse(min_version):
        raise RuntimeError(
            f"Package {package} {installed_version} is too old! At least {min_version} is required."
        )

# For ipython sessions, activate awaitless
try:
    from . import awaitless
    awaitless.load_ipython_extension()
except (ImportError, RuntimeError):
    pass

# Alias from long.internal.name down to something reasonable
from .core.crs import (
    CRS,
    Crate,
    ReadoutChannel,
    ReadoutModule,
    Resonator,
    Wafer,
)

from .core.hardware_map import macro, algorithm, HardwareMap
from .core.dirfile import load_dirfile
from .core.session import load_session, get_session, set_session, load_from_database

from . import core, algorithms, streamer, tools

from .tuber import TuberRemoteError

# vim: sts=4 ts=4 sw=4 tw=78 smarttab expandtab



"""
Mock Configuration Dialog for Periscope (Unified SoT)
=====================================================

This dialog exposes only the parameters defined in rfmux.mock.config
(Single Source of Truth). It returns a configuration dict that can be passed
directly to MockCRS.generate_resonators().

Groups:
- Basic: number of resonances, frequency range (GHz), random seed
- Bias: automatic bias enable and bias amplitude (normalized)
- Advanced (collapsible):
  - Physics: T (K), Popt (W)
  - Circuit: Lg (H), Cc (F), L_junk (H), C_variation, Cc_variation
  - Readout: Vin (V), input_atten_dB, system_termination (Ω), ZLNA (Ω), GLNA (dB)
  - Noise: nqp_noise_enabled, nqp_noise_std_factor, udp_noise_level (counts)
  - Convergence: convergence_tolerance
"""

from .utils import QtWidgets, QtCore, QtGui
import re
import math
import numpy as np
from rfmux.mock import config as mc
from rfmux.mr_resonator.mr_complex_resonator import MR_complex_resonator
from rfmux.mr_resonator.mr_lekid import MR_LEKID
from . import settings
from .custom_material_dialog import CustomMaterialDialog, ManageCustomMaterialsDialog


class ScientificDoubleValidator(QtGui.QDoubleValidator):
    """Validator that accepts scientific notation."""
    def validate(self, string, pos):
        if re.match(r'^-?(\d+\.?\d*|\d*\.\d+)([eE][+-]?\d+)?$', string) or string == '' or string == '-':
            return (QtGui.QValidator.State.Acceptable, string, pos)
        return (QtGui.QValidator.State.Invalid, string, pos)


class MockConfigurationDialog(QtWidgets.QDialog):
    """
    Dialog for configuring mock CRS simulation parameters based on SoT (mock_config).
    """

    def __init__(self, parent=None, current_config=None):
        super().__init__(parent)
        self.setWindowTitle("Mock Mode Configuration")
        self.setModal(True)
        self.setMinimumWidth(1000)

        # Store current configuration or use defaults
        self.current_config = current_config or {}

        # Create UI
        self._create_ui()

        # Load current or default values
        self._load_current_values()

        # Connect signals for interactions
        self._connect_signals()

        # Trigger initial derived parameter update
        self._update_all_derived()

    def _connect_signals(self):
        """Connect signals for live update of derived parameters."""
        # Physics mode inputs
        for widget in [self.T_edit, self.Popt_edit, self.width_edit, 
                       self.thickness_edit, self.length_edit]:
            widget.textChanged.connect(self._update_all_derived)
        
        # Circuit design
        for widget in [self.Lg_edit, self.L_junk_edit, self.Cc_edit,
                       self.Vin_edit, self.input_atten_edit, self.ZLNA_edit]:
            widget.textChanged.connect(self._update_all_derived)
            
        self.freq_start_spin.valueChanged.connect(self._update_all_derived)

    def _create_ui(self):
        layout = QtWidgets.QVBoxLayout(self)

        # Warning message with theme-appropriate background
        warning_label = QtWidgets.QLabel(
            "⚠️ <b>Periscope is running in MOCK MODE</b><br>"
            "This simulates a CRS board with virtual KID resonators."
        )
        palette = self.palette()
        bg_color = palette.color(QtGui.QPalette.ColorRole.Window)
        is_dark = bg_color.lightness() < 128
        if is_dark:
            style = "QLabel { background-color: #664422; color: #FFE4B5; padding: 10px; border-radius: 5px; }"
        else:
            style = "QLabel { background-color: #FFF4E0; color: #663300; padding: 10px; border-radius: 5px; }"
        warning_label.setStyleSheet(style)
        layout.addWidget(warning_label)

        layout.addSpacing(8)

        # Basic and Bias parameters
        layout.addWidget(self._create_basic_group())
        layout.addSpacing(8)
        layout.addWidget(self._create_bias_group())

        layout.addSpacing(8)

        # Advanced parameters (collapsible)
        self.advanced_widget = self._create_advanced_widget()
        layout.addWidget(self.advanced_widget)

        layout.addStretch()

        # Buttons row
        button_layout = QtWidgets.QHBoxLayout()
        self.reset_button = QtWidgets.QPushButton("Reset to Defaults")
        self.reset_button.clicked.connect(self._reset_to_defaults)
        button_layout.addWidget(self.reset_button)
        button_layout.addStretch()
        button_box = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.StandardButton.Ok | QtWidgets.QDialogButtonBox.StandardButton.Cancel
        )
        button_box.accepted.connect(self._validate_and_accept)
        button_box.rejected.connect(self.reject)
        button_layout.addWidget(button_box)
        layout.addLayout(button_layout)

    def _create_basic_group(self) -> QtWidgets.QGroupBox:
        group = QtWidgets.QGroupBox("Basic")
        layout = QtWidgets.QGridLayout(group)
        row = 0

        # Number of resonances
        layout.addWidget(QtWidgets.QLabel("Number of Resonances:"), row, 0)
        self.num_resonances_spin = QtWidgets.QSpinBox()
        self.num_resonances_spin.setRange(1, 10000)
        self.num_resonances_spin.setSingleStep(1)
        self.num_resonances_spin.setToolTip("Total number of resonances to generate")
        layout.addWidget(self.num_resonances_spin, row, 1)

        # Frequency start/end (GHz)
        layout.addWidget(QtWidgets.QLabel("Frequency Start (GHz):"), row, 2)
        self.freq_start_spin = QtWidgets.QDoubleSpinBox()
        self.freq_start_spin.setRange(0.001, 100.0)
        self.freq_start_spin.setSingleStep(0.001)
        self.freq_start_spin.setDecimals(3)
        self.freq_start_spin.setToolTip("Start frequency for resonance generation [GHz]")
        layout.addWidget(self.freq_start_spin, row, 3)

        row += 1
        layout.addWidget(QtWidgets.QLabel("Frequency End (GHz):"), row, 0)
        self.freq_end_spin = QtWidgets.QDoubleSpinBox()
        self.freq_end_spin.setRange(0.001, 100.0)
        self.freq_end_spin.setSingleStep(0.001)
        self.freq_end_spin.setDecimals(3)
        self.freq_end_spin.setToolTip("End frequency for resonance generation [GHz]")
        layout.addWidget(self.freq_end_spin, row, 1)

        # Random seed (optional)
        layout.addWidget(QtWidgets.QLabel("Random Seed:"), row, 2)
        self.random_seed_edit = QtWidgets.QLineEdit()
        self.random_seed_edit.setValidator(QtGui.QIntValidator(0, 2**31 - 1, self.random_seed_edit))
        self.random_seed_edit.setToolTip("Set a seed for reproducible layout (leave blank for random)")
        layout.addWidget(self.random_seed_edit, row, 3)

        return group

    def _create_bias_group(self) -> QtWidgets.QGroupBox:
        group = QtWidgets.QGroupBox("Bias")
        layout = QtWidgets.QGridLayout(group)
        row = 0

        self.auto_bias_check = QtWidgets.QCheckBox("Automatically bias KIDs")
        self.auto_bias_check.setToolTip("Configure tones on generated resonator frequencies")
        layout.addWidget(self.auto_bias_check, row, 0, 1, 2)

        layout.addWidget(QtWidgets.QLabel("Bias Amplitude (normalized):"), row, 2)
        self.bias_amplitude_spin = QtWidgets.QDoubleSpinBox()
        self.bias_amplitude_spin.setRange(0.0001, 1.0)
        self.bias_amplitude_spin.setSingleStep(0.001)
        self.bias_amplitude_spin.setDecimals(4)
        self.bias_amplitude_spin.setToolTip("Amplitude for automatic biasing (e.g., 0.01 ≈ -40 dBm)")
        layout.addWidget(self.bias_amplitude_spin, row, 3)

        self.auto_bias_check.toggled.connect(self.bias_amplitude_spin.setEnabled)
        return group

    def _create_advanced_widget(self) -> QtWidgets.QWidget:
        widget = QtWidgets.QWidget()
        vbox = QtWidgets.QVBoxLayout(widget)
        vbox.setContentsMargins(0, 0, 0, 0)

        self.advanced_toggle = QtWidgets.QPushButton("▶ Advanced Parameters")
        self.advanced_toggle.setCheckable(True)
        self.advanced_toggle.setFlat(True)
        self.advanced_toggle.setStyleSheet("QPushButton { text-align: left; padding: 5px; }")
        self.advanced_toggle.setToolTip("Show or hide advanced physics, circuit, readout, noise, and solver settings.")
        self.advanced_toggle.clicked.connect(self._toggle_advanced)
        vbox.addWidget(self.advanced_toggle)

        self.advanced_container = QtWidgets.QWidget()
        self.advanced_container.setVisible(False)
        grid = QtWidgets.QHBoxLayout(self.advanced_container)

        # Left column
        left = QtWidgets.QVBoxLayout()
        left.addWidget(self._create_mkids_group())
        left.addStretch()

        # Right column
        right = QtWidgets.QVBoxLayout()
        right.addWidget(self._create_readout_group())
        right.addWidget(self._create_pulse_injection_group())
        right.addWidget(self._create_noise_group())
        right.addWidget(self._create_simulation_realism_group())
        right.addStretch()

        grid.addLayout(left)
        grid.addLayout(right)
        vbox.addWidget(self.advanced_container)
        return widget

    def _toggle_advanced(self):
        vis = self.advanced_container.isVisible()
        self.advanced_container.setVisible(not vis)
        self.advanced_toggle.setText("▼ Advanced Parameters" if not vis else "▶ Advanced Parameters")

    def _create_mkids_group(self) -> QtWidgets.QGroupBox:
        """Create unified MKIDs section with physics-driven parameters."""
        group = QtWidgets.QGroupBox("MKIDs (Microwave Kinetic Inductance Detectors)")
        layout = QtWidgets.QVBoxLayout(group)
        
        # Physics-driven content
        physics_page = self._create_physics_driven_page()
        layout.addWidget(physics_page)
        
        # Circuit Design Parameters (shared between both modes, always visible)
        circuit_group = QtWidgets.QGroupBox("Circuit Design Parameters")
        circuit_layout = QtWidgets.QGridLayout(circuit_group)
        
        circuit_layout.addWidget(QtWidgets.QLabel("Lg (geometric, nH):"), 0, 0)
        self.Lg_edit = QtWidgets.QLineEdit()
        self.Lg_edit.setValidator(ScientificDoubleValidator())
        self.Lg_edit.setToolTip("Geometric inductance in nanohenries. Fixed by fabrication geometry.")
        circuit_layout.addWidget(self.Lg_edit, 0, 1)
        
        circuit_layout.addWidget(QtWidgets.QLabel("L_junk (nH):"), 0, 2)
        self.L_junk_edit = QtWidgets.QLineEdit()
        self.L_junk_edit.setValidator(ScientificDoubleValidator())
        self.L_junk_edit.setToolTip("Parasitic inductance in nH. Dilutes α_k = Lk/(Lk+Lg+L_junk).")
        circuit_layout.addWidget(self.L_junk_edit, 0, 3)
        
        circuit_layout.addWidget(QtWidgets.QLabel("Cc (fF):"), 1, 0)
        self.Cc_edit = QtWidgets.QLineEdit()
        self.Cc_edit.setValidator(ScientificDoubleValidator())
        self.Cc_edit.setToolTip("Coupling capacitor in femtofarads. Controls coupling strength.")
        circuit_layout.addWidget(self.Cc_edit, 1, 1)
        
        circuit_layout.addWidget(QtWidgets.QLabel("C variation:"), 1, 2)
        self.C_variation_edit = QtWidgets.QLineEdit()
        self.C_variation_edit.setValidator(ScientificDoubleValidator())
        self.C_variation_edit.setToolTip("Fractional σ of capacitance variation (e.g., 0.01 = 1%).")
        circuit_layout.addWidget(self.C_variation_edit, 1, 3)
        
        circuit_layout.addWidget(QtWidgets.QLabel("Cc variation:"), 2, 0)
        self.Cc_variation_edit = QtWidgets.QLineEdit()
        self.Cc_variation_edit.setValidator(ScientificDoubleValidator())
        self.Cc_variation_edit.setToolTip("Fractional σ of coupling capacitor variation.")
        circuit_layout.addWidget(self.Cc_variation_edit, 2, 1)
        
        layout.addWidget(circuit_group)
        
        # Derived Properties - combined two-column layout
        derived_group = QtWidgets.QGroupBox("Derived Properties (live-updated)")
        derived_main_layout = QtWidgets.QHBoxLayout(derived_group)
        
        # Left column - physics derived values
        left_form = QtWidgets.QFormLayout()
        left_form.setFieldGrowthPolicy(QtWidgets.QFormLayout.FieldGrowthPolicy.ExpandingFieldsGrow)
        
        self.nqp_label = QtWidgets.QLabel("—")
        self.nqp_label.setStyleSheet("color: #1E90FF; font-family: monospace;")
        self.nqp_label.setToolTip("Quasiparticle density. Increases with T and Popt.")
        left_form.addRow("nqp (µm⁻³):", self.nqp_label)
        
        self.sigma1_label = QtWidgets.QLabel("—")
        self.sigma1_label.setStyleSheet("color: #1E90FF; font-family: monospace;")
        self.sigma1_label.setToolTip("Real part of complex conductivity (related to loss).")
        left_form.addRow("σ₁ (S/m):", self.sigma1_label)
        
        self.sigma2_label = QtWidgets.QLabel("—")
        self.sigma2_label.setStyleSheet("color: #1E90FF; font-family: monospace;")
        self.sigma2_label.setToolTip("Imaginary part of complex conductivity (related to kinetic inductance).")
        left_form.addRow("σ₂ (S/m):", self.sigma2_label)
        
        self.lk_square_label = QtWidgets.QLabel("—")
        self.lk_square_label.setStyleSheet("color: #1E90FF; font-family: monospace;")
        self.lk_square_label.setToolTip("Kinetic inductance per square. Material property at given T, Popt.")
        left_form.addRow("Lk□ (pH/□):", self.lk_square_label)
        
        self.lk_label = QtWidgets.QLabel("—")
        self.lk_label.setStyleSheet("color: #1E90FF; font-family: monospace;")
        self.lk_label.setToolTip("Total kinetic inductance = Lk□ × (length/width).")
        left_form.addRow("Lk (nH):", self.lk_label)
        
        self.r_label = QtWidgets.QLabel("—")
        self.r_label.setStyleSheet("color: #1E90FF; font-family: monospace;")
        self.r_label.setToolTip("Total series resistance scaled from surface resistance.")
        left_form.addRow("R (mΩ):", self.r_label)
        
        # Right column - circuit derived values
        right_form = QtWidgets.QFormLayout()
        right_form.setFieldGrowthPolicy(QtWidgets.QFormLayout.FieldGrowthPolicy.ExpandingFieldsGrow)
        
        self.ltotal_label = QtWidgets.QLabel("—")
        self.ltotal_label.setStyleSheet("color: #1E90FF; font-family: monospace;")
        self.ltotal_label.setToolTip("Total inductance = Lk + Lg + L_junk. Determines resonance frequency.")
        right_form.addRow("L_total (nH):", self.ltotal_label)
        
        self.alpha_k_label = QtWidgets.QLabel("—")
        self.alpha_k_label.setStyleSheet("color: #1E90FF; font-family: monospace;")
        self.alpha_k_label.setToolTip("Kinetic inductance fraction = Lk / L_total. Determines responsivity.")
        right_form.addRow("α_k:", self.alpha_k_label)
        
        self.qr_label = QtWidgets.QLabel("—")
        self.qr_label.setStyleSheet("color: #1E90FF; font-family: monospace;")
        self.qr_label.setToolTip("Loaded quality factor. Limited by both internal loss and coupling.")
        right_form.addRow("Q_r (loaded):", self.qr_label)
        
        self.qi_label = QtWidgets.QLabel("—")
        self.qi_label.setStyleSheet("color: #1E90FF; font-family: monospace;")
        self.qi_label.setToolTip("Internal quality factor. Limited by loss only.")
        right_form.addRow("Q_i (internal):", self.qi_label)
        
        self.qc_label = QtWidgets.QLabel("—")
        self.qc_label.setStyleSheet("color: #1E90FF; font-family: monospace;")
        self.qc_label.setToolTip("Coupling quality factor. Higher Qc = weaker coupling.")
        right_form.addRow("Q_c (coupling):", self.qc_label)
        
        # Add columns to main layout
        derived_main_layout.addLayout(left_form)
        derived_main_layout.addLayout(right_form)
        
        layout.addWidget(derived_group)
        
        return group
    
    def _create_physics_driven_page(self) -> QtWidgets.QWidget:
        """Create physics-driven mode page with geometry inputs and derived properties."""
        page = QtWidgets.QWidget()
        layout = QtWidgets.QVBoxLayout(page)
        
        # Material & Operating Point subsection
        mat_group = QtWidgets.QGroupBox("Material and Operating Point")
        mat_layout = QtWidgets.QGridLayout(mat_group)
        
        # Material combo with manage button
        mat_layout.addWidget(QtWidgets.QLabel("Material:"), 0, 0)
        
        material_h_layout = QtWidgets.QHBoxLayout()
        self.material_combo = QtWidgets.QComboBox()
        self.material_combo.setToolTip("Superconductor material. Determines Tc, N₀, τ₀, σN constants.")
        self.material_combo.currentTextChanged.connect(self._on_material_changed)
        material_h_layout.addWidget(self.material_combo, 1)
        
        # Manage materials button
        self.manage_materials_btn = QtWidgets.QPushButton("⚙ Manage...")
        self.manage_materials_btn.setToolTip("Add, edit, or delete custom materials")
        self.manage_materials_btn.clicked.connect(self._manage_materials)
        self.manage_materials_btn.setMaximumWidth(100)
        material_h_layout.addWidget(self.manage_materials_btn)
        
        mat_layout.addLayout(material_h_layout, 0, 1, 1, 3)
        
        # Material constants label
        self.mat_constants_label = QtWidgets.QLabel("Tc=1.2K, N₀=1.72×10¹⁰ µm⁻³eV⁻¹, τ₀=438ns")
        self.mat_constants_label.setStyleSheet("color: gray; font-size: 9pt;")
        self.mat_constants_label.setToolTip(
            "<b>Material Constants:</b><br>"
            "• <b>Tc</b> = Critical temperature<br>"
            "• <b>N₀</b> = Density of states at Fermi level<br>"
            "• <b>τ₀</b> = Quasiparticle recombination time at Tc"
        )
        mat_layout.addWidget(self.mat_constants_label, 1, 0, 1, 4)
        
        # T and Popt
        mat_layout.addWidget(QtWidgets.QLabel("Temperature (K):"), 2, 0)
        self.T_edit = QtWidgets.QLineEdit()
        self.T_edit.setValidator(ScientificDoubleValidator())
        self.T_edit.setToolTip("Operating temperature in Kelvin. Lower T → lower nqp → higher Lk → higher Q.")
        mat_layout.addWidget(self.T_edit, 2, 1)
        
        mat_layout.addWidget(QtWidgets.QLabel("Optical Power (W):"), 2, 2)
        self.Popt_edit = QtWidgets.QLineEdit()
        self.Popt_edit.setValidator(ScientificDoubleValidator())
        self.Popt_edit.setToolTip("Incident optical power in Watts. Higher Popt → higher nqp → lower Q.")
        mat_layout.addWidget(self.Popt_edit, 2, 3)
        
        layout.addWidget(mat_group)
        
        # Load materials into combo box
        self._refresh_material_list()
        
        # Geometry subsection
        geom_group = QtWidgets.QGroupBox("Geometry")
        geom_layout = QtWidgets.QGridLayout(geom_group)
        
        # Width, thickness, length inputs
        geom_layout.addWidget(QtWidgets.QLabel("Width (µm):"), 0, 0)
        self.width_edit = QtWidgets.QLineEdit("2.0")
        self.width_edit.setValidator(ScientificDoubleValidator())
        self.width_edit.setToolTip("Strip width in micrometers. Affects number of squares.")
        geom_layout.addWidget(self.width_edit, 0, 1)
        
        geom_layout.addWidget(QtWidgets.QLabel("Thickness (nm):"), 0, 2)
        self.thickness_edit = QtWidgets.QLineEdit("30")
        self.thickness_edit.setValidator(ScientificDoubleValidator())
        self.thickness_edit.setToolTip("Film thickness in nanometers. Thicker films reduce surface impedance.")
        geom_layout.addWidget(self.thickness_edit, 0, 3)
        
        geom_layout.addWidget(QtWidgets.QLabel("Length (µm):"), 1, 0)
        self.length_edit = QtWidgets.QLineEdit("9000")
        self.length_edit.setValidator(ScientificDoubleValidator())
        self.length_edit.setToolTip("Strip length in micrometers. Longer → higher Lk and R.")
        geom_layout.addWidget(self.length_edit, 1, 1)
        
        # Volume display (calculated)
        self.volume_label = QtWidgets.QLabel("Volume: 540 µm³")
        self.volume_label.setStyleSheet("color: #1E90FF; font-weight: bold;")
        geom_layout.addWidget(self.volume_label, 1, 2, 1, 2)
        
        layout.addWidget(geom_group)
        
        return page
    
    def _refresh_material_list(self):
        """Populate material combo with built-in + custom materials."""
        current = self.material_combo.currentText()
        self.material_combo.blockSignals(True)  # Prevent triggering _on_material_changed during refresh
        self.material_combo.clear()
        
        # Built-in materials
        self.material_combo.addItem("Al (Aluminum)")
        
        # Custom materials
        customs = settings.get_custom_materials()
        for name in sorted(customs.keys()):
            self.material_combo.addItem(name)
        
        # Restore selection or default to Al
        idx = self.material_combo.findText(current)
        if idx >= 0:
            self.material_combo.setCurrentIndex(idx)
        else:
            self.material_combo.setCurrentIndex(0)
        
        self.material_combo.blockSignals(False)
        # Trigger update for current selection
        self._on_material_changed(self.material_combo.currentText())
    
    def _on_material_changed(self, material_text: str):
        """Update material constants label when material changes."""
        if not material_text:
            return
        
        # Extract material name (remove description if present)
        mat_name = material_text.split()[0] if material_text else "Al"
        
        try:
            props = settings.get_material_properties(mat_name)
            Tc = props['Tc']
            N0 = props['N0']
            tau0_ns = props['tau0'] * 1e9  # s to ns
            
            # Display Rs and thickness_ref if available
            Rs = props.get('Rs', None)
            thickness_ref = props.get('thickness_ref', None)
            
            if Rs is not None and thickness_ref is not None:
                self.mat_constants_label.setText(
                    f"Tc={Tc:.3g}K, N₀={N0:.3g} µm⁻³eV⁻¹, τ₀={tau0_ns:.3g}ns, "
                    f"Rs={Rs:.3g}Ω/□ @{thickness_ref:.0f}nm"
                )
            else:
                # Legacy materials without Rs
                self.mat_constants_label.setText(
                    f"Tc={Tc:.3g}K, N₀={N0:.3g} µm⁻³eV⁻¹, τ₀={tau0_ns:.3g}ns"
                )
        except ValueError:
            self.mat_constants_label.setText("Material properties not found")
        
        # Trigger recalculation
        self._update_all_derived()
    
    def _manage_materials(self):
        """Open dialog to manage custom materials."""
        dialog = ManageCustomMaterialsDialog(parent=self)
        if dialog.exec() == QtWidgets.QDialog.DialogCode.Accepted:
            # Refresh material list after dialog closes
            self._refresh_material_list()
    
    def _update_all_derived(self):
        """Update all derived parameters based on physics-driven inputs."""
        try:
            # Physics mode: calculate from T, Popt, geometry
            try:
                T = float(self.T_edit.text())
                Popt = float(self.Popt_edit.text())
                width_um = float(self.width_edit.text())
                thickness_nm = float(self.thickness_edit.text())
                length_um = float(self.length_edit.text())
            except ValueError:
                return
            
            # Convert to SI
            width = width_um * 1e-6
            thickness = thickness_nm * 1e-9
            length = length_um * 1e-6
            
            # Calculate volume
            volume_um3 = width_um * (thickness_nm / 1000) * length_um
            self.volume_label.setText(f"Volume: {volume_um3:.1f} µm³")
            
            # Get frequency
            freq = self.freq_start_spin.value() * 1e9
            if freq <= 0:
                return
            
            try:
                Lg = float(self.Lg_edit.text()) * 1e-9
                L_junk = float(self.L_junk_edit.text()) * 1e-9
                Cc = float(self.Cc_edit.text()) * 1e-15
                Vin = float(self.Vin_edit.text())
                input_atten_dB = float(self.input_atten_edit.text())
                ZLNA = float(self.ZLNA_edit.text())
            except ValueError:
                return
            
            # Get material properties
            mat_text = self.material_combo.currentText()
            mat_name = mat_text.split()[0] if mat_text else "Al"
            try:
                mat_props = settings.get_material_properties(mat_name)
            except ValueError:
                # If material not found, use Al defaults
                mat_props = settings.get_material_properties("Al")
            
            # Create reference resonator with geometry and custom material properties
            # MR_complex_resonator now accepts explicit N0, tau0, Tc, sigmaN parameters
            # N0 unit conversion: mat_props['N0'] is in µm⁻³eV⁻¹, internal uses µm⁻³J⁻¹
            # Correct conversion: N0_J⁻¹ = N0_eV⁻¹ / (1.602e-19 J/eV)
            ref_params = {
                'T': T, 'Popt': Popt, 'C': 1e-12, 'Cc': Cc, 'fix_Lg': Lg,
                'L_junk': L_junk, 'Vin': Vin, 'input_atten_dB': input_atten_dB,
                'base_readout_f': freq, 'verbose': False,
                'ZLNA': complex(ZLNA, 0), 'GLNA': 1.0,
                'width': width, 'thickness': thickness, 'length': length,
                # Material properties - pass directly (no post-hoc override needed)
                'Tc': mat_props['Tc'],
                'N0': mat_props['N0'] / 1.602e-19,  # Convert µm⁻³eV⁻¹ → µm⁻³J⁻¹
                'tau0': mat_props['tau0'],
                'sigmaN': mat_props['sigmaN']
            }
            ref_res = MR_complex_resonator(**ref_params)
            
            # Extract intermediate values
            nqp = ref_res.calc_nqp()
            sigma1 = ref_res.sigma1_dark
            sigma2 = ref_res.sigma2_dark
            Lk = ref_res.lekid.Lk
            R = ref_res.lekid.R
            
            # Calculate Lk per square
            Lk_square = Lk / (length / width) if (length / width) > 0 else 0
            
            # Update derived property labels
            self.nqp_label.setText(f"{nqp:.3e}")
            self.sigma1_label.setText(f"{sigma1:.3e}")
            self.sigma2_label.setText(f"{sigma2:.3e}")
            self.lk_square_label.setText(f"{Lk_square*1e12:.2f}")
            self.lk_label.setText(f"{Lk*1e9:.3f}")
            self.r_label.setText(f"{R*1e6:.3f}")
            
            # Calculate final derived properties using already-parsed Lg, L_junk, Cc, etc.
            # (No second read needed - we already have these values from above)
            L_total = Lk + Lg + L_junk
            alpha_k = Lk / L_total if L_total > 0 else 0
            
            # Calculate Q values
            freq = self.freq_start_spin.value() * 1e9
            if freq > 0:
                C = 1.0 / ((2 * np.pi * freq)**2 * L_total) if L_total > 0 else 1e-12
                
                lekid = MR_LEKID(
                    C=C, Lk=Lk, Lg=Lg, R=R, Cc=Cc, 
                    L_junk=L_junk, Vin=Vin, 
                    system_termination=50.0, input_atten_dB=input_atten_dB,
                    ZLNA=complex(ZLNA, 0)
                )
                
                Qr, Qi, Qc = lekid.compute_Q_values()
            else:
                Qr = Qi = Qc = 0
            
            # Update final labels
            self.ltotal_label.setText(f"{L_total*1e9:.2f}")
            self.alpha_k_label.setText(f"{alpha_k:.4f}")
            self.qr_label.setText(f"{Qr:,.0f}")
            self.qi_label.setText(f"{Qi:,.0f}")
            self.qc_label.setText(f"{Qc:,.0f}")
            
            # Debug: print to console to verify values are updating
            # print(f"DEBUG _update_all_derived: Lg={Lg*1e9:.2f}nH, L_junk={L_junk*1e9:.2f}nH, Lk={Lk*1e9:.3f}nH -> L_total={L_total*1e9:.2f}nH, α_k={alpha_k:.4f}")
            
        except Exception as e:
            # Log the exception for debugging (was silently swallowed)
            print(f"DEBUG _update_all_derived exception: {type(e).__name__}: {e}")

    def _create_readout_group(self) -> QtWidgets.QGroupBox:
        group = QtWidgets.QGroupBox("Readout")
        layout = QtWidgets.QGridLayout(group)
        row = 0

        layout.addWidget(QtWidgets.QLabel("Vin (V):"), row, 0)
        self.Vin_edit = QtWidgets.QLineEdit()
        self.Vin_edit.setValidator(ScientificDoubleValidator())
        self.Vin_edit.setToolTip("Input drive voltage at the source before attenuation (Volts).")
        layout.addWidget(self.Vin_edit, row, 1)

        layout.addWidget(QtWidgets.QLabel("Input Attenuation (dB):"), row, 2)
        self.input_atten_edit = QtWidgets.QLineEdit()
        self.input_atten_edit.setValidator(ScientificDoubleValidator())
        self.input_atten_edit.setToolTip("Input attenuation applied to Vin (dB).")
        layout.addWidget(self.input_atten_edit, row, 3)

        row += 1
        layout.addWidget(QtWidgets.QLabel("System Termination (Ω):"), row, 0)
        self.system_termination_edit = QtWidgets.QLineEdit()
        self.system_termination_edit.setValidator(ScientificDoubleValidator())
        self.system_termination_edit.setToolTip("Readout line termination impedance (Ohms).")
        layout.addWidget(self.system_termination_edit, row, 1)

        layout.addWidget(QtWidgets.QLabel("ZLNA (Ω):"), row, 2)
        self.ZLNA_edit = QtWidgets.QLineEdit()
        self.ZLNA_edit.setValidator(ScientificDoubleValidator())
        self.ZLNA_edit.setToolTip("LNA input impedance (Ohms). Typically 50 Ω.")
        layout.addWidget(self.ZLNA_edit, row, 3)

        row += 1
        layout.addWidget(QtWidgets.QLabel("GLNA (dB):"), row, 0)
        self.GLNA_db_spin = QtWidgets.QDoubleSpinBox()
        self.GLNA_db_spin.setRange(-60.0, 60.0)
        self.GLNA_db_spin.setSingleStep(0.1)
        self.GLNA_db_spin.setDecimals(1)
        self.GLNA_db_spin.setToolTip("Voltage gain of the LNA in dB (20·log10(Vout/Vin)). Internally converted to linear V/V.")
        layout.addWidget(self.GLNA_db_spin, row, 1)

        return group

    def _create_noise_group(self) -> QtWidgets.QGroupBox:
        group = QtWidgets.QGroupBox("Noise")
        layout = QtWidgets.QGridLayout(group)
        row = 0

        self.nqp_noise_enabled_cb = QtWidgets.QCheckBox("Enable QP noise")
        self.nqp_noise_enabled_cb.setToolTip("Enable random fluctuations of quasiparticle density per evaluation.")
        layout.addWidget(self.nqp_noise_enabled_cb, row, 0, 1, 2)

        layout.addWidget(QtWidgets.QLabel("Quasiparticle Noise (frac):"), row, 2)
        self.nqp_noise_std_edit = QtWidgets.QLineEdit()
        self.nqp_noise_std_edit.setValidator(ScientificDoubleValidator())
        self.nqp_noise_std_edit.setToolTip("Std dev as a fraction of base quasiparticle density (e.g., 0.001 = 0.1%).")
        layout.addWidget(self.nqp_noise_std_edit, row, 3)

        row += 1
        layout.addWidget(QtWidgets.QLabel("ADC Noise (counts):"), row, 0)
        self.udp_noise_edit = QtWidgets.QLineEdit()
        self.udp_noise_edit.setValidator(ScientificDoubleValidator())
        self.udp_noise_edit.setToolTip("Additive ADC noise level in counts")
        layout.addWidget(self.udp_noise_edit, row, 1)

        return group

    def _create_pulse_injection_group(self) -> QtWidgets.QGroupBox:
        group = QtWidgets.QGroupBox("Pulse Injection")
        layout = QtWidgets.QGridLayout(group)
        row = 0

        layout.addWidget(QtWidgets.QLabel("Period (s):"), row, 0)
        self.pulse_period_edit = QtWidgets.QLineEdit()
        self.pulse_period_edit.setValidator(ScientificDoubleValidator())
        self.pulse_period_edit.setToolTip("Pulse period in seconds (used in periodic mode).")
        layout.addWidget(self.pulse_period_edit, row, 1)

        layout.addWidget(QtWidgets.QLabel("Probability (/s):"), row, 2)
        self.pulse_probability_edit = QtWidgets.QLineEdit()
        self.pulse_probability_edit.setValidator(ScientificDoubleValidator())
        self.pulse_probability_edit.setToolTip("Per-resonator per-second probability (random mode).\nEffective per-update chance ≈ probability × dt.")
        layout.addWidget(self.pulse_probability_edit, row, 3)

        row += 1
        layout.addWidget(QtWidgets.QLabel("Tau rise (s):"), row, 0)
        self.pulse_tau_rise_edit = QtWidgets.QLineEdit()
        self.pulse_tau_rise_edit.setValidator(ScientificDoubleValidator())
        self.pulse_tau_rise_edit.setToolTip("Exponential rise time constant (seconds).")
        layout.addWidget(self.pulse_tau_rise_edit, row, 1)

        layout.addWidget(QtWidgets.QLabel("Tau decay (s):"), row, 2)
        self.pulse_tau_decay_edit = QtWidgets.QLineEdit()
        self.pulse_tau_decay_edit.setValidator(ScientificDoubleValidator())
        self.pulse_tau_decay_edit.setToolTip("Exponential decay time constant (seconds).")
        layout.addWidget(self.pulse_tau_decay_edit, row, 3)

        row += 1
        layout.addWidget(QtWidgets.QLabel("Amplitude (× base nqp):"), row, 0)
        self.pulse_amplitude_edit = QtWidgets.QLineEdit()
        self.pulse_amplitude_edit.setValidator(ScientificDoubleValidator())
        self.pulse_amplitude_edit.setToolTip("Multiplicative factor relative to base quasiparticle density.\nExample: 2.0 doubles base nqp.")
        layout.addWidget(self.pulse_amplitude_edit, row, 1)

        layout.addWidget(QtWidgets.QLabel("Resonators:"), row, 2)
        self.pulse_resonators_edit = QtWidgets.QLineEdit()
        self.pulse_resonators_edit.setToolTip('Target resonators: "all" or CSV of 0-based indices (e.g., 0,1,7).')
        layout.addWidget(self.pulse_resonators_edit, row, 3)

        # Random amplitude distribution (random mode)
        row += 1
        layout.addWidget(QtWidgets.QLabel("Random amplitude mode:"), row, 0)
        self.random_amp_mode_combo = QtWidgets.QComboBox()
        self.random_amp_mode_combo.addItems(["fixed", "uniform", "lognormal"])
        self.random_amp_mode_combo.setToolTip('Random pulse amplitude distribution in "random" mode.')
        layout.addWidget(self.random_amp_mode_combo, row, 1)

        row += 1
        layout.addWidget(QtWidgets.QLabel("Uniform min:"), row, 0)
        self.random_amp_min_edit = QtWidgets.QLineEdit()
        self.random_amp_min_edit.setValidator(ScientificDoubleValidator())
        self.random_amp_min_edit.setToolTip("Minimum amplitude (× base nqp) for uniform distribution (≥ 1.0).")
        layout.addWidget(self.random_amp_min_edit, row, 1)

        layout.addWidget(QtWidgets.QLabel("Uniform max:"), row, 2)
        self.random_amp_max_edit = QtWidgets.QLineEdit()
        self.random_amp_max_edit.setValidator(ScientificDoubleValidator())
        self.random_amp_max_edit.setToolTip("Maximum amplitude (× base nqp) for uniform distribution (≥ min).")
        layout.addWidget(self.random_amp_max_edit, row, 3)

        row += 1
        layout.addWidget(QtWidgets.QLabel("Lognormal mean (μ):"), row, 0)
        self.random_amp_logmean_edit = QtWidgets.QLineEdit()
        self.random_amp_logmean_edit.setValidator(ScientificDoubleValidator())
        self.random_amp_logmean_edit.setToolTip("Mean parameter μ for lognormal amplitude distribution.")
        layout.addWidget(self.random_amp_logmean_edit, row, 1)

        layout.addWidget(QtWidgets.QLabel("Lognormal sigma (σ):"), row, 2)
        self.random_amp_logsigma_edit = QtWidgets.QLineEdit()
        self.random_amp_logsigma_edit.setValidator(ScientificDoubleValidator())
        self.random_amp_logsigma_edit.setToolTip("Sigma parameter σ (≥ 0) for lognormal amplitude distribution.")
        layout.addWidget(self.random_amp_logsigma_edit, row, 3)

        # Connect mode change to toggling of fields
        self.random_amp_mode_combo.currentTextChanged.connect(self._on_random_amp_mode_changed)

        return group

    def _on_random_amp_mode_changed(self, mode: str):
        # Enable/disable distribution-specific fields
        mode = (mode or "").lower()
        is_uniform = mode == "uniform"
        is_lognormal = mode == "lognormal"
        # Uniform fields
        for w in (self.random_amp_min_edit, self.random_amp_max_edit):
            w.setEnabled(is_uniform)
        # Lognormal fields
        for w in (self.random_amp_logmean_edit, self.random_amp_logsigma_edit):
            w.setEnabled(is_lognormal)

    def _create_simulation_realism_group(self) -> QtWidgets.QGroupBox:
        group = QtWidgets.QGroupBox("Simulation Realism")
        layout = QtWidgets.QGridLayout(group)

        # Convergence tolerance (solver accuracy)
        layout.addWidget(QtWidgets.QLabel("I^2 Convergence tolerance:"), 0, 0)
        self.conv_tol_edit = QtWidgets.QLineEdit()
        self.conv_tol_edit.setValidator(ScientificDoubleValidator())
        self.conv_tol_edit.setToolTip("Solver accuracy: lower = more accurate, slower (e.g., 1e-9 default).")
        layout.addWidget(self.conv_tol_edit, 0, 1)

        # Cache QP step (fraction)
        layout.addWidget(QtWidgets.QLabel("Cache QP step (frac):"), 1, 0)
        self.cache_qp_step_edit = QtWidgets.QLineEdit()
        self.cache_qp_step_edit.setValidator(ScientificDoubleValidator())
        self.cache_qp_step_edit.setToolTip("QP quantization as fraction of base QP (e.g., 0.001 = 0.1%). Larger = coarser = more reuse.")
        layout.addWidget(self.cache_qp_step_edit, 1, 1)

        return group

    def _reset_to_defaults(self):
        cfg = mc.defaults()
        # Basic
        self.num_resonances_spin.setValue(int(cfg["num_resonances"]))
        self.freq_start_spin.setValue(float(cfg["freq_start"]) / 1e9)
        self.freq_end_spin.setValue(float(cfg["freq_end"]) / 1e9)
        self.random_seed_edit.setText("" if cfg["resonator_random_seed"] is None else str(cfg["resonator_random_seed"]))
        # Bias
        self.auto_bias_check.setChecked(bool(cfg["auto_bias_kids"]))
        self.bias_amplitude_spin.setValue(float(cfg["bias_amplitude"]))
        # Physics-driven mode
        self.T_edit.setText(str(cfg["T"]))
        self.Popt_edit.setText(str(cfg["Popt"]))
        self.width_edit.setText(str(round(cfg.get("width", 2e-6) * 1e6, 6)))  # m to µm, round to 6 decimals
        self.thickness_edit.setText(str(round(cfg.get("thickness", 30e-9) * 1e9, 6)))  # m to nm, round to 6 decimals
        self.length_edit.setText(str(round(cfg.get("length", 9000e-6) * 1e6, 6)))  # m to µm, round to 6 decimals
        # Circuit
        self.Lg_edit.setText(str(cfg["Lg"] * 1e9))  # H to nH
        self.Cc_edit.setText(str(cfg["Cc"] * 1e15))  # F to fF
        self.L_junk_edit.setText(str(cfg["L_junk"] * 1e9))  # H to nH
        self.C_variation_edit.setText(str(cfg["C_variation"]))
        self.Cc_variation_edit.setText(str(cfg["Cc_variation"]))
        # Readout
        self.Vin_edit.setText(str(cfg["Vin"]))
        self.input_atten_edit.setText(str(cfg["input_atten_dB"]))
        self.system_termination_edit.setText(str(cfg["system_termination"]))
        self.ZLNA_edit.setText(str(cfg["ZLNA"]))
        try:
            glna_val = float(cfg["GLNA"])
            glna_db = 20.0 * math.log10(glna_val) if glna_val > 0 else 0.0
        except Exception:
            glna_db = 0.0
        self.GLNA_db_spin.setValue(glna_db)
        # Noise
        self.nqp_noise_enabled_cb.setChecked(bool(cfg["nqp_noise_enabled"]))
        self.nqp_noise_std_edit.setText(str(cfg["nqp_noise_std_factor"]))
        self.udp_noise_edit.setText(str(cfg["udp_noise_level"]))
        # Physics Realism & Cache
        self.conv_tol_edit.setText(str(cfg["convergence_tolerance"]))
        self.cache_qp_step_edit.setText(str(cfg["cache_qp_step"]))

        # QP Pulses
        self.pulse_period_edit.setText(str(cfg["pulse_period"]))
        self.pulse_probability_edit.setText(str(cfg["pulse_probability"]))
        self.pulse_tau_rise_edit.setText(str(cfg["pulse_tau_rise"]))
        self.pulse_tau_decay_edit.setText(str(cfg["pulse_tau_decay"]))
        self.pulse_amplitude_edit.setText(str(cfg["pulse_amplitude"]))
        resonators = cfg.get("pulse_resonators", "all")
        if isinstance(resonators, list):
            self.pulse_resonators_edit.setText(",".join(str(int(r)) for r in resonators))
        else:
            self.pulse_resonators_edit.setText(str(resonators))

        # Random amplitude distribution
        ram = str(cfg.get("pulse_random_amp_mode"))
        idx_ram = max(0, self.random_amp_mode_combo.findText(ram))
        self.random_amp_mode_combo.setCurrentIndex(idx_ram)
        self.random_amp_min_edit.setText(str(cfg.get("pulse_random_amp_min")))
        self.random_amp_max_edit.setText(str(cfg.get("pulse_random_amp_max")))
        self.random_amp_logmean_edit.setText(str(cfg.get("pulse_random_amp_logmean")))
        self.random_amp_logsigma_edit.setText(str(cfg.get("pulse_random_amp_logsigma")))
        self._on_random_amp_mode_changed(self.random_amp_mode_combo.currentText())
        
        # Trigger derived parameter update
        self._update_all_derived()

    def _load_current_values(self):
        if not self.current_config:
            self._reset_to_defaults()
            return
        cfg = {**mc.defaults(), **self.current_config}  # fill missing with defaults

        # Basic
        self.num_resonances_spin.setValue(int(cfg.get("num_resonances")))
        self.freq_start_spin.setValue(float(cfg.get("freq_start")) / 1e9)
        self.freq_end_spin.setValue(float(cfg.get("freq_end")) / 1e9)
        seed = cfg.get("resonator_random_seed", None)
        self.random_seed_edit.setText("" if seed in (None, "") else str(int(seed)))

        # Bias
        self.auto_bias_check.setChecked(bool(cfg.get("auto_bias_kids")))
        self.bias_amplitude_spin.setValue(float(cfg.get("bias_amplitude")))

        # Physics-driven mode
        self.T_edit.setText(str(cfg.get("T")))
        self.Popt_edit.setText(str(cfg.get("Popt")))
        self.width_edit.setText(str(round(cfg.get("width", 2e-6) * 1e6, 6)))  # m to µm, round to 6 decimals
        self.thickness_edit.setText(str(round(cfg.get("thickness", 30e-9) * 1e9, 6)))  # m to nm, round to 6 decimals
        self.length_edit.setText(str(round(cfg.get("length", 9000e-6) * 1e6, 6)))  # m to µm, round to 6 decimals
        
        # Material selection (if present in config)
        if 'material' in cfg:
            mat_name = cfg['material']
            # Find material in combo (might be "Al (Aluminum)" or just "Al" or custom "Nb")
            idx = -1
            for i in range(self.material_combo.count()):
                combo_text = self.material_combo.itemText(i)
                combo_mat = combo_text.split()[0]
                if combo_mat == mat_name:
                    idx = i
                    break
            if idx >= 0:
                self.material_combo.setCurrentIndex(idx)

        # Circuit
        self.Lg_edit.setText(str(cfg.get("Lg") * 1e9))  # H to nH
        self.Cc_edit.setText(str(cfg.get("Cc") * 1e15))  # F to fF
        self.L_junk_edit.setText(str(cfg.get("L_junk") * 1e9))  # H to nH
        self.C_variation_edit.setText(str(cfg.get("C_variation")))
        self.Cc_variation_edit.setText(str(cfg.get("Cc_variation")))

        # Readout
        self.Vin_edit.setText(str(cfg.get("Vin")))
        self.input_atten_edit.setText(str(cfg.get("input_atten_dB")))
        self.system_termination_edit.setText(str(cfg.get("system_termination")))
        self.ZLNA_edit.setText(str(cfg.get("ZLNA")))
        try:
            glna_val = float(cfg.get("GLNA"))
            glna_db = 20.0 * math.log10(glna_val) if glna_val > 0 else 0.0
        except Exception:
            glna_db = 0.0
        self.GLNA_db_spin.setValue(glna_db)

        # Noise
        self.nqp_noise_enabled_cb.setChecked(bool(cfg.get("nqp_noise_enabled")))
        self.nqp_noise_std_edit.setText(str(cfg.get("nqp_noise_std_factor")))
        self.udp_noise_edit.setText(str(cfg.get("udp_noise_level")))

        # Physics Realism & Cache
        self.conv_tol_edit.setText(str(cfg.get("convergence_tolerance")))
        self.cache_qp_step_edit.setText(str(cfg.get("cache_qp_step")))

        # QP Pulses
        self.pulse_period_edit.setText(str(cfg.get("pulse_period")))
        self.pulse_probability_edit.setText(str(cfg.get("pulse_probability")))
        self.pulse_tau_rise_edit.setText(str(cfg.get("pulse_tau_rise")))
        self.pulse_tau_decay_edit.setText(str(cfg.get("pulse_tau_decay")))
        self.pulse_amplitude_edit.setText(str(cfg.get("pulse_amplitude")))
        resonators = cfg.get("pulse_resonators", "all")
        if isinstance(resonators, list):
            self.pulse_resonators_edit.setText(",".join(str(int(r)) for r in resonators))
        else:
            self.pulse_resonators_edit.setText(str(resonators))

        # Random amplitude distribution
        ram = str(cfg.get("pulse_random_amp_mode"))
        idx_ram = max(0, self.random_amp_mode_combo.findText(ram))
        self.random_amp_mode_combo.setCurrentIndex(idx_ram)
        self.random_amp_min_edit.setText(str(cfg.get("pulse_random_amp_min")))
        self.random_amp_max_edit.setText(str(cfg.get("pulse_random_amp_max")))
        self.random_amp_logmean_edit.setText(str(cfg.get("pulse_random_amp_logmean")))
        self.random_amp_logsigma_edit.setText(str(cfg.get("pulse_random_amp_logsigma")))
        self._on_random_amp_mode_changed(self.random_amp_mode_combo.currentText())

    def _validate_and_accept(self):
        # Frequency range
        if self.freq_start_spin.value() >= self.freq_end_spin.value():
            QtWidgets.QMessageBox.warning(self, "Invalid Input", "Frequency Start must be less than Frequency End.")
            return

        # Validate numeric text fields that must parse
        try:
            # Physics-driven mode
            float(self.T_edit.text()); float(self.Popt_edit.text())
            float(self.width_edit.text()); float(self.thickness_edit.text()); float(self.length_edit.text())
            # Circuit
            float(self.Lg_edit.text()); float(self.Cc_edit.text()); float(self.L_junk_edit.text())
            float(self.C_variation_edit.text()); float(self.Cc_variation_edit.text())
            # Readout
            float(self.Vin_edit.text()); float(self.input_atten_edit.text())
            float(self.system_termination_edit.text()); float(self.ZLNA_edit.text()); float(self.GLNA_db_spin.value())
            # Noise and physics realism
            float(self.nqp_noise_std_edit.text()); float(self.udp_noise_edit.text()); float(self.conv_tol_edit.text())
            float(self.cache_qp_step_edit.text())
            # QP Pulses
            float(self.pulse_period_edit.text())
            float(self.pulse_probability_edit.text())
            float(self.pulse_tau_rise_edit.text())
            float(self.pulse_tau_decay_edit.text())
            float(self.pulse_amplitude_edit.text())
            # Random distribution fields (parse regardless; defaults present)
            float(self.random_amp_min_edit.text()); float(self.random_amp_max_edit.text())
            float(self.random_amp_logmean_edit.text()); float(self.random_amp_logsigma_edit.text())
        except ValueError:
            QtWidgets.QMessageBox.warning(self, "Invalid Input", "Please enter valid numeric values for all fields.")
            return

        self.accept()

    def get_configuration(self) -> dict:
        """
        Return a dictionary of all parameter values matching config.MOCK_DEFAULTS keys.
        """
        seed_text = self.random_seed_edit.text().strip()
        seed_value = int(seed_text) if seed_text != "" else None

        config_dict = {
            # Basic
            "num_resonances": int(self.num_resonances_spin.value()),
            "freq_start": float(self.freq_start_spin.value()) * 1e9,  # GHz -> Hz
            "freq_end": float(self.freq_end_spin.value()) * 1e9,      # GHz -> Hz
            "resonator_random_seed": seed_value,

            # Bias
            "auto_bias_kids": bool(self.auto_bias_check.isChecked()),
            "bias_amplitude": float(self.bias_amplitude_spin.value()),

            # Physics
            "T": float(self.T_edit.text()),
            "Popt": float(self.Popt_edit.text()),

            # Circuit
            "Lg": float(self.Lg_edit.text()) * 1e-9,  # nH -> H
            "Cc": float(self.Cc_edit.text()) * 1e-15,  # fF -> F
            "L_junk": float(self.L_junk_edit.text()) * 1e-9,  # nH -> H
            "C_variation": float(self.C_variation_edit.text()),
            "Cc_variation": float(self.Cc_variation_edit.text()),

            # Readout
            "Vin": float(self.Vin_edit.text()),
            "input_atten_dB": float(self.input_atten_edit.text()),
            "system_termination": float(self.system_termination_edit.text()),
            "ZLNA": float(self.ZLNA_edit.text()),
            "GLNA": float(10 ** (self.GLNA_db_spin.value() / 20.0)),

            # Noise
            "nqp_noise_enabled": bool(self.nqp_noise_enabled_cb.isChecked()),
            "nqp_noise_std_factor": float(self.nqp_noise_std_edit.text()),
            "udp_noise_level": float(self.udp_noise_edit.text()),

            # Physics Realism & Cache
            "convergence_tolerance": float(self.conv_tol_edit.text()),
            "cache_qp_step": float(self.cache_qp_step_edit.text()),

            # QP Pulses
            "pulse_period": float(self.pulse_period_edit.text()),
            "pulse_probability": float(self.pulse_probability_edit.text()),
            "pulse_tau_rise": float(self.pulse_tau_rise_edit.text()),
            "pulse_tau_decay": float(self.pulse_tau_decay_edit.text()),
            "pulse_amplitude": float(self.pulse_amplitude_edit.text()),
            "pulse_resonators": (lambda txt: "all" if txt.strip().lower() == "all" or txt.strip()=="" else [
                int(s) for s in [p.strip() for p in txt.split(",")] if s.isdigit()
            ])(self.pulse_resonators_edit.text()),

            # Random amplitude distribution (random mode)
            "pulse_random_amp_mode": str(self.random_amp_mode_combo.currentText()),
            "pulse_random_amp_min": float(self.random_amp_min_edit.text()),
            "pulse_random_amp_max": float(self.random_amp_max_edit.text()),
            "pulse_random_amp_logmean": float(self.random_amp_logmean_edit.text()),
            "pulse_random_amp_logsigma": float(self.random_amp_logsigma_edit.text()),
        }
        
        # Add geometry parameters
        config_dict['width'] = float(self.width_edit.text()) * 1e-6  # µm -> m
        config_dict['thickness'] = float(self.thickness_edit.text()) * 1e-9  # nm -> m
        config_dict['length'] = float(self.length_edit.text()) * 1e-6  # µm -> m
        
        # Add material name
        mat_text = self.material_combo.currentText()
        mat_name = mat_text.split()[0] if mat_text else "Al"
        config_dict['material'] = mat_name
        
        # If custom material, include properties for reproducibility
        if mat_name != "Al":
            try:
                props = settings.get_material_properties(mat_name)
                config_dict['material_Tc'] = props['Tc']
                config_dict['material_N0'] = props['N0']
                config_dict['material_tau0'] = props['tau0']
                if 'sigmaN' in props:
                    config_dict['material_sigmaN'] = props['sigmaN']
            except ValueError:
                pass
        
        return config_dict



"""
bias_kids: A measurement algorithm for biasing KIDs at their optimal operating points
based on multisweep characterization data.
"""

import numpy as np
import asyncio
import warnings
from typing import Union, Dict, List, Optional, Any, Tuple, Callable
from scipy.signal import butter, filtfilt



def bandpass_filter(data: np.ndarray, fs: float, lowcut: float, highcut: float, order: int = 4) -> np.ndarray:
    """
    Apply a bandpass filter to the data.
    
    Args:
        data: Input signal
        fs: Sampling frequency (Hz)
        lowcut: Low frequency cutoff (Hz)
        highcut: High frequency cutoff (Hz)
        order: Filter order (default: 4)
        
    Returns:
        Filtered signal
    """
    nyq = 0.5 * fs
    b, a = butter(order, [lowcut / nyq, highcut / nyq], btype='band')
    return filtfilt(b, a, data)


async def find_optimal_phases_parallel(
    crs,
    bias_configs: Dict[int, Dict],
    module: int,
    num_samples: int = 300,
    apply_bandpass: bool = True,
    fs: float = 597,
    lowcut: float = 5,
    highcut: float = 20,
    phase_step: int = 5
) -> Dict[int, Tuple[float, float]]:
    """
    Find optimal ADC phases for multiple channels in parallel.
    
    Args:
        crs: CRS object
        bias_configs: Dictionary of bias configurations {det_idx: config}
        module: Module number
        num_samples: Number of samples to collect at each phase
        apply_bandpass: Whether to apply bandpass filter to Q data
        fs: Sampling frequency (Hz) - only used if apply_bandpass is True
        lowcut: Bandpass filter low cutoff (Hz) - only used if apply_bandpass is True
        highcut: Bandpass filter high cutoff (Hz) - only used if apply_bandpass is True
        phase_step: Phase step size in degrees
        
    Returns:
        Dictionary of {det_idx: (best_phase_degrees, max_std_q)}
    """
    optimal_phases = {}
    
    # Initialize best phases and stds for each detector
    for det_idx in bias_configs:
        optimal_phases[det_idx] = (0.0, -np.inf)
    
    # Scan through phases
    for phase in range(0, 360, phase_step):
        # Set phase for all channels simultaneously
        async with crs.tuber_context() as ctx:
            for det_idx, config in bias_configs.items():
                ctx.set_phase(phase, units=crs.UNITS.DEGREES, target=crs.TARGET.ADC, 
                             channel=config['channel'], module=module)
            await ctx()
        
        # Collect samples for all channels at once
        samples = await crs.get_samples(num_samples, channel=None, module=module, average=False)
        
        # Process each detector's Q data
        for det_idx, config in bias_configs.items():
            channel_idx = config['channel'] - 1  # Convert to 0-based index
            
            # Extract Q data for this channel
            q_data = np.array(samples.q[channel_idx])
            
            # Calculate standard deviation (with or without bandpass filter)
            try:
                if apply_bandpass:
                    q_processed = bandpass_filter(q_data, fs, lowcut, highcut)
                else:
                    q_processed = q_data
                    
                std_q = float(np.std(q_processed))
                
                # Update if this is better than the current best
                current_best_phase, current_best_std = optimal_phases[det_idx]
                if std_q > current_best_std:
                    optimal_phases[det_idx] = (float(phase), std_q)
                    
            except Exception as e:
                if apply_bandpass:
                    warnings.warn(f"Bandpass filter failed for detector {det_idx} at phase {phase}°: {e}")
                else:
                    warnings.warn(f"Failed to process Q data for detector {det_idx} at phase {phase}°: {e}")
                continue
    
    # Report results
    filter_desc = "filtered " if apply_bandpass else ""
    for det_idx, (best_phase, best_std) in optimal_phases.items():
        print(f"Detector {det_idx}: Optimal phase = {best_phase}°, {filter_desc}std(Q) = {best_std:.4f}")
    
    return optimal_phases


def _extract_data_from_gui_format(gui_results: Dict) -> Tuple[Optional[Dict[Tuple[float, str], Dict[int, Dict]]], Dict]:
    """
    Extract detector data from GUI multisweep results format.
    
    Args:
        gui_results: Dictionary with 'results_by_iteration' key
        
    Returns:
        Tuple of (multiamp_data, metadata)
        where multiamp_data is {(amplitude, direction): {detector_idx: data}}
        Returns (None, {}) if not GUI format
    """
    if 'results_by_iteration' not in gui_results:
        # Not GUI format, return None to indicate not applicable
        return None, {}
        
    multiamp_data = {}
    metadata = {
        'iterations': [],
        'amplitudes': set(),
        'directions': set()
    }
    
    for iteration_data in gui_results['results_by_iteration']:
        iteration = iteration_data.get('iteration')
        amplitude = iteration_data.get('amplitude')
        direction = iteration_data.get('direction', 'upward')
        data = iteration_data.get('data', {})
        
        key = (amplitude, direction)
        multiamp_data[key] = data
        
        metadata['iterations'].append({
            'iteration': iteration,
            'amplitude': amplitude,
            'direction': direction
        })
        metadata['amplitudes'].add(amplitude)
        metadata['directions'].add(direction)
    
    return multiamp_data, metadata


async def bias_kids(
    crs,
    multisweep_results: Union[Dict, List[Dict]],
    nonlinear_threshold: float = 0.77,
    fallback_to_lowest: bool = True,
    optimize_phase: bool = False,
    bandpass_params: Optional[Dict[str, float]] = None,
    num_phase_samples: int = 300,
    phase_step: int = 5,
    *,
    module: Optional[Union[int, List[int]]] = None,
    progress_callback: Optional[Callable] = None
) -> Union[Dict[int, Dict], List[Dict[int, Dict]]]:
    """
    Bias KIDs at their optimal operating points based on multisweep characterization.
    
    This algorithm analyzes multisweep results to find the best amplitude for each
    detector (highest amplitude that is not bifurcated and has nonlinear parameter < threshold),
    then programs the detectors with the appropriate frequency, phase, and amplitude.
    
    Args:
        crs: The CRS object to use for hardware communication.
        multisweep_results: Can be one of:
                           - Dict with 'results_by_iteration' key: Multi-amplitude GUI format
                           - Dict[int, Dict]: Single amplitude results
                           - List[Dict]: Multiple modules
        nonlinear_threshold (float): Maximum acceptable nonlinear parameter 'a'.
                                   Defaults to 0.77.
        fallback_to_lowest (bool): If True and no suitable amplitude found,
                                 use the lowest available amplitude. If False,
                                 skip the detector. Defaults to True.
        optimize_phase (bool): If True, scan through ADC phases to find the phase that
                             maximizes variance in bandpass-filtered Q timestream.
                             Defaults to False.
        bandpass_params (dict, optional): Parameters for bandpass filter used in phase optimization.
                                        Keys: 'lowcut' (Hz), 'highcut' (Hz), 'fs' (sampling freq Hz).
                                        Defaults: {'lowcut': 5, 'highcut': 20, 'fs': 597}.
        num_phase_samples (int): Number of samples to collect for phase optimization.
                               Defaults to 300.
        phase_step (int): Phase step size in degrees for optimization scan.
                        Defaults to 5.
        module (int | list[int], optional): Target module(s). If None, extracted from results.
        progress_callback (callable, optional): Function called with (module, progress_percentage).
        
    Returns:
        Dictionary or list of dictionaries containing only the biased detectors' data.
        Each entry includes the original multisweep data plus:
        - 'bias_channel': The assigned channel number (1-based)
        - 'bias_amplitude': The amplitude selected for biasing
        - 'bifurcation_suspected': Whether any amplitude showed bifurcation
        - 'bias_successful': Whether the detector was successfully biased
        - 'optimal_phase_degrees': The optimal ADC phase found (if phase optimization enabled)
        - 'phase_optimization_std': The bandpass-filtered Q std at optimal phase
    """
    
    # Check if this is GUI format with multiple amplitudes
    if isinstance(multisweep_results, dict) and 'results_by_iteration' in multisweep_results:
        # Extract multi-amplitude data from GUI format
        multiamp_data, metadata = _extract_data_from_gui_format(multisweep_results)
        
        if multiamp_data is not None:
            # Find optimal configuration for each detector
            optimal_configs = analyze_multiamp_data(multiamp_data, nonlinear_threshold, fallback_to_lowest)
            
            # Convert to single result set using optimal configurations
            single_results = {}
            for det_idx, config in optimal_configs.items():
                single_results[det_idx] = config['selected_data'].copy()
                # Add metadata about amplitude selection
                single_results[det_idx]['selected_amplitude'] = config['selected_amplitude']
                single_results[det_idx]['bifurcation_ever_seen'] = config.get('bifurcation_ever_seen', False)
                
            # Proceed with single-amplitude logic using optimal results
            multisweep_results = single_results
        
    # Handle list of results (multiple modules)
    elif isinstance(multisweep_results, list):
        if module is None:
            # Extract module numbers from the data if possible
            # This would require the multisweep results to contain module info
            # For now, assume modules are sequential starting from 1
            module = list(range(1, len(multisweep_results) + 1))
        elif not isinstance(module, list):
            module = [module]
            
        if len(multisweep_results) != len(module):
            raise ValueError(f"Number of result sets ({len(multisweep_results)}) "
                           f"doesn't match number of modules ({len(module)})")
        
        # Process each module
        tasks = []
        for mod_idx, (mod_results, mod_num) in enumerate(zip(multisweep_results, module)):
            tasks.append(
                bias_kids(
                    crs=crs,
                    multisweep_results=mod_results,
                    nonlinear_threshold=nonlinear_threshold,
                    fallback_to_lowest=fallback_to_lowest,
                    optimize_phase=optimize_phase,
                    bandpass_params=bandpass_params,
                    num_phase_samples=num_phase_samples,
                    phase_step=phase_step,
                    module=mod_num,
                    progress_callback=progress_callback
                )
            )
        
        return await asyncio.gather(*tasks)
    
    # Single module processing
    if module is None:
        raise ValueError("Module must be specified for single result set")
    
    # Get current NCO frequency
    nco_freq = await crs.get_nco_frequency(module=module)
    
    # Base frequency (Nyquist frequency) for quantization
    base_freq = 298.0232238769531  # Hz
    
    # Set default bandpass parameters if not provided
    if bandpass_params is None:
        bandpass_params = {'lowcut': 5, 'highcut': 20, 'fs': 597}
    
    # Analyze each detector to find optimal bias point
    bias_configs = {}
    total_detectors = len(multisweep_results)
    
    for det_idx, det_data in multisweep_results.items():
        # Check if this is multi-amplitude data
        # Multi-amplitude data would need to be organized differently
        # For now, assume single amplitude per detector
        
        # Extract key parameters
        is_bifurcated = det_data.get('is_bifurcated', False)
        nonlinear_params = det_data.get('nonlinear_fit_params', {})
        nonlinear_a = nonlinear_params.get('a', float('inf'))
        
        # Check if detector meets criteria
        suitable = not is_bifurcated and nonlinear_a < nonlinear_threshold
        
        if suitable or fallback_to_lowest:
            # Prepare bias configuration
            bias_freq = det_data.get('bias_frequency', det_data.get('original_center_frequency'))
            sweep_amp = det_data.get('sweep_amplitude')
            
            if bias_freq is None or sweep_amp is None:
                warnings.warn(f"Detector {det_idx}: Missing bias frequency or amplitude")
                continue
                
            # Channel assignment: det_idx is already 1-based from multisweep
            channel = det_idx
            
            # Quantize the absolute bias frequency to nearest multiple of base frequency
            quantized_bias_freq = round(bias_freq / base_freq) * base_freq
            
            # Calculate channel frequency relative to NCO
            channel_freq = quantized_bias_freq - nco_freq
            
            bias_configs[det_idx] = {
                'channel': int(channel),  # Ensure it's a Python int
                'frequency': float(channel_freq),  # Channel frequency relative to NCO
                'original_bias_frequency': float(bias_freq),  # Store original for reference
                'quantized_bias_frequency': float(quantized_bias_freq),  # Quantized absolute frequency
                'amplitude': float(sweep_amp),  # Ensure it's a Python float
                'phase': 0.0,  # No phase rotation applied anymore
                'suitable': suitable,
                'bifurcation_suspected': is_bifurcated,
                'nonlinear_a': nonlinear_a
            }
            
            if not suitable:
                warnings.warn(f"Detector {det_idx}: No suitable amplitude found "
                            f"(bifurcated={is_bifurcated}, a={nonlinear_a:.3f}). "
                            f"Using fallback amplitude.")
    
    # # Clear all channels first
    # max_channels = 1024
    # async with crs.tuber_context() as ctx:
    #     for ch in range(1, max_channels + 1):
    #         ctx.set_amplitude(0, channel=ch, module=module)
    #     await ctx()
    
    # Program the selected detectors
    successfully_biased = {}
    
    # First, set up all tones without phase optimization
    async with crs.tuber_context() as ctx:
        for det_idx, config in bias_configs.items():
            try:
                ctx.set_frequency(config['frequency'], channel=config['channel'], module=module)
                ctx.set_amplitude(config['amplitude'], channel=config['channel'], module=module)
                ctx.set_phase(config['phase'], units=crs.UNITS.DEGREES, target=crs.TARGET.ADC, channel=config['channel'], module=module)
            except Exception as e:
                print(f"[Bias] Failed to set up detector {det_idx}: {e}")
                continue
        await ctx()
    
    # Now perform phase optimization if requested
    if optimize_phase:
        print(f"Optimizing phases for {len(bias_configs)} detectors in parallel...")
        
        # Determine if bandpass filter should be applied
        # If bandpass_params contains 'apply_bandpass', use that value, otherwise default to True
        apply_bandpass = True
        if bandpass_params is not None:
            apply_bandpass = bool(bandpass_params.get('apply_bandpass', True))
        
        # Find optimal phases for all detectors in parallel
        optimal_phases = await find_optimal_phases_parallel(
            crs=crs,
            bias_configs=bias_configs,
            module=module,  # type: ignore  # module is guaranteed to be int at this point
            num_samples=num_phase_samples,
            apply_bandpass=apply_bandpass,
            fs=bandpass_params.get('fs', 597) if bandpass_params else 597,
            lowcut=bandpass_params.get('lowcut', 5) if bandpass_params else 5,
            highcut=bandpass_params.get('highcut', 20) if bandpass_params else 20,
            phase_step=phase_step
        )
    else:
        # No optimization - all phases are 0
        optimal_phases = {det_idx: (0.0, None) for det_idx in bias_configs}
    
    # Apply the optimal phases and create result data
    for det_idx, config in bias_configs.items():
        try:
            optimal_phase, phase_std = optimal_phases.get(det_idx, (0.0, None))
            
            # Set the optimal phase for this channel
            if optimal_phase != 0.0:
                await crs.set_phase(optimal_phase, crs.UNITS.DEGREES, crs.TARGET.ADC, 
                                   channel=config['channel'], module=module)
            
            # Copy the original multisweep data and add bias info
            biased_data = multisweep_results[det_idx].copy()
            biased_data['bias_channel'] = config['channel']
            biased_data['bifurcation_suspected'] = config['bifurcation_suspected']
            biased_data['bias_successful'] = True
            biased_data['optimal_phase_degrees'] = optimal_phase
            biased_data['phase_optimization_std'] = phase_std
            
            # If we have df calibration and applied a phase, rotate it
            if 'df_calibration' in biased_data and biased_data['df_calibration'] is not None and optimal_phase != 0.0:
                # Rotate the df calibration by the applied phase
                phase_rad = np.radians(optimal_phase)
                rotation_factor = np.exp(1j * phase_rad)
                biased_data['df_calibration'] *= rotation_factor
                biased_data['df_calibration_rotated'] = True
            
            successfully_biased[det_idx] = biased_data
            
        except Exception as e:
            warnings.warn(f"Failed to bias detector {det_idx}: {e}")
    
    # Progress callback
    if progress_callback:
        progress_callback(module, 100.0)
    
    # Log summary
    print(f"Module {module}: Successfully biased {len(successfully_biased)}/{total_detectors} detectors")
    
    return successfully_biased


def analyze_multiamp_data(
    multiamp_data: Dict[Tuple[float, str], Dict[int, Dict]],
    nonlinear_threshold: float = 0.77,
    fallback_to_lowest: bool = True
) -> Dict[int, Dict[str, Any]]:
    """
    Analyze multi-amplitude multisweep results to find optimal bias points.
    
    Args:
        multiamp_data: Dictionary with (amplitude, direction) tuple as key, detector data as value
        nonlinear_threshold: Maximum acceptable nonlinear parameter
        fallback_to_lowest: Whether to use lowest amplitude as fallback
        
    Returns:
        Dictionary with detector index as key, optimal configuration as value
    """
    optimal_configs = {}
    
    # Get all unique amplitudes and sort them
    amplitudes = sorted(set(amp for amp, _ in multiamp_data.keys()), reverse=True)
    
    # Get all detector indices from the first entry
    first_data = next(iter(multiamp_data.values()))
    detector_indices = list(first_data.keys())
    
    for det_idx in detector_indices:
        # Collect data across all amplitudes for this detector
        amp_analysis = []
        bifurcation_ever_seen = False
        
        for amp in amplitudes:
            # Check both directions if available
            for direction in ['upward', 'downward']:
                key = (amp, direction)
                if key not in multiamp_data:
                    continue
                    
                if det_idx not in multiamp_data[key]:
                    continue
                    
                det_data = multiamp_data[key][det_idx]
                is_bifurcated = det_data.get('is_bifurcated', False)
                nonlinear_params = det_data.get('nonlinear_fit_params', {})
                nonlinear_a = nonlinear_params.get('a', float('inf'))
                nonlinear_success = det_data.get('nonlinear_fit_success', False)
                
                if is_bifurcated:
                    bifurcation_ever_seen = True
                
                # Only consider nonlinear parameter if fit was successful
                if nonlinear_success:
                    suitable = not is_bifurcated and nonlinear_a < nonlinear_threshold
                else:
                    suitable = not is_bifurcated
                    
                amp_analysis.append({
                    'amplitude': amp,
                    'direction': direction,
                    'is_bifurcated': is_bifurcated,
                    'nonlinear_a': nonlinear_a,
                    'nonlinear_fit_success': nonlinear_success,
                    'suitable': suitable,
                    'data': det_data
                })
        
        # Find the highest suitable amplitude
        suitable_entries = [a for a in amp_analysis if a['suitable']]
        
        if suitable_entries:
            # Use the first suitable (which is highest amplitude due to sort)
            optimal = suitable_entries[0]
        elif fallback_to_lowest and amp_analysis:
            # No suitable amplitude found, use lowest amplitude
            # Sort by amplitude (ascending) to get lowest
            sorted_by_amp = sorted(amp_analysis, key=lambda x: x['amplitude'])
            optimal = sorted_by_amp[0]
        else:
            optimal = None
            
        if optimal:
            optimal_configs[det_idx] = {
                'selected_amplitude': optimal['amplitude'],
                'selected_direction': optimal['direction'],
                'selected_data': optimal['data'],
                'bifurcation_ever_seen': bifurcation_ever_seen,
                'all_amplitudes': amp_analysis
            }
    
    return optimal_configs



from collections.abc import Sequence, Mapping
from collections import namedtuple
import sys

try:
    import numpy

    have_numpy = True
except ImportError:
    have_numpy = False

# Prefer SimpleJSON, but fall back on built-in
try:
    import simplejson as json
except ImportError:
    import json  # type: ignore[no-redef]

try:
    import orjson

    have_orjson = True
except ImportError:
    have_orjson = False

try:
    import cbor2

    have_cbor = True
except ImportError:
    have_cbor = False


class TuberResult:
    def __init__(self, d):
        "Allow dotted accessors, like an object"
        self.__dict__.update(d)

    def __iter__(self):
        "Make the results object iterate as a list of keys, like a dict"
        return iter(self.__dict__)

    def __repr__(self):
        "Return a concise representation string"
        return repr(self.__dict__)

    def _todict(self):
        """Recursively convert TuberResult object to a dictionary."""
        result = {}
        for key, value in self.__dict__.items():
            if isinstance(value, TuberResult):
                # Recursively convert nested TuberResult
                result[key] = value._todict()
            elif isinstance(value, (list, tuple)):
                # Convert any TuberResult objects in sequences
                converted = []
                for item in value:
                    if isinstance(item, TuberResult):
                        converted.append(item._todict())
                    else:
                        converted.append(item)
                result[key] = type(value)(converted)  # Preserve list/tuple type
            elif isinstance(value, dict):
                # Convert any TuberResult objects in dictionaries
                converted = {}
                for k, v in value.items():
                    if isinstance(v, TuberResult):
                        converted[k] = v._todict()
                    else:
                        converted[k] = v
                result[key] = converted
            else:
                # Keep other types as-is
                result[key] = value
        return result


def wrap_bytes_for_json(obj):
    """
    JSON cannot (natively) encode bytes, so we provide a simple encoding for them.
    This allows uniformity when using either JSON or binary formats (CBOR, etc.)
    which do have native binary support. The JSON encoding is not meant to be
    especially efficient, since anyone wanting seriously move around significant
    amounts of binary data should use another format, but it provides a
    consistent, readable/debuggable, fall-back.
    """
    if isinstance(obj, bytes):
        data = [int(v) for v in obj]
        return {"bytes": data}
    return obj


def cbor_encode_ndarray(enc, arr):
    # At the moment, this handles only contiguous arrays of data types which can be represented
    # as CBOR typed arrays, as these can be handled with a singleblock copy of the underlying data,
    # with no per-element handling.

    # start with big endian tags, and then patch up later if the data turn out to be little endian
    type_tags = {
        "u": {
            1: 64,
            2: 65,
            4: 66,
            8: 67,
        },
        "i": {
            1: 72,
            2: 73,
            4: 74,
            8: 75,
        },
        "f": {
            2: 80,
            4: 81,
            8: 82,
            16: 83,
        },
    }
    if arr.dtype.kind not in type_tags or arr.dtype.itemsize not in type_tags[arr.dtype.kind]:
        raise cbor2.CBOREncodeTypeError(
            f"Serialization of numpy arrays with element type {arr.dtype} is not implemented"
        )
    type_tag = type_tags[arr.dtype.kind][arr.dtype.itemsize]
    # add 4 to type tag if little endian if sizeof(type) > 1
    if arr.dtype.itemsize > 1 and (
        arr.dtype.byteorder == "<" or (arr.dtype.byteorder == "=" and sys.byteorder == "little")
    ):
        type_tag += 4

    if arr.flags.c_contiguous:
        md_tag = 40  # row-major
    elif arr.flags.f_contiguous:
        md_tag = 1040  # column-major
    else:
        raise cbor2.CBOREncodeTypeError("Serialization of non-contiguous numpy arrays is not implemented")

    enc.encode_length(6, md_tag)  # multi-dimensional array header, a tag (type 6) of the correct type
    enc.encode_length(4, 2)  # payload of the m-d array is always an array (type 4) of length 2
    enc.encode_length(4, len(arr.shape))  # the first item in the outer array is the array of extents
    for extent in arr.shape:
        enc.encode_int(extent)
    # the second item in the outer array is the array entries, for which we use a typed array
    enc.encode_length(6, type_tag)
    # the typed array payload is a bytestring (type 2)
    enc.encode_length(2, arr.nbytes)
    # call write directly on the stream object to avoid unnecessary copies
    enc.fp.write(arr.data)


def cbor_augment_encode(enc, obj):
    if isinstance(obj, numpy.ndarray):
        cbor_encode_ndarray(enc, obj)
        return
    raise cbor2.CBOREncodeTypeError(f"Unsupported object for CBOR encoding {type(obj)}")


def cbor_tag_decode(dec, tag):
    if have_numpy and tag.tag >= 64 and tag.tag <= 87 and tag.tag != 76:  # Typed arrays
        is_float = tag.tag & 0x10
        is_signed = tag.tag & 0x8
        is_le = tag.tag & 0x4
        ll = tag.tag & 0x3
        element_size = 1 << ll
        if is_float:  # floats are one power of two larger
            element_size <<= 1
        # due to the cap of 87 on the tag, we will never see invalid 'signed' float combinations
        dt = numpy.dtype(f"{'<' if is_le else '>'}{'f' if is_float else 'i' if is_signed else 'u'}{element_size}")
        if len(tag.value) % element_size != 0:
            raise cbor2.CBORDecodeValueError(
                f"Invalid data size ({len(tag.value)}) for typed array with tag {tag.tag}, interpreted as {dt}"
            )
        # create a 1-D, row-major array to contain all of the data, which can have more detailed
        # shape and ordering information applied later
        arr = numpy.zeros(len(tag.value) // element_size, dtype=dt, order="C")
        # splat the data into the array's memory
        arr.data.cast("B")[:] = tag.value
        return arr
    if have_numpy and (tag.tag == 40 or tag.tag == 1040):
        if not isinstance(tag.value, Sequence):
            raise cbor2.CBORDecodeValueError(f"Invalid raw data for multi-dimensional array tag ({tag.tag})")
        if len(tag.value) != 2:
            raise cbor2.CBORDecodeValueError(f"Invalid raw array length for multi-dimensional array tag ({tag.tag})")
        if not isinstance(tag.value[0], Sequence) or not isinstance(tag.value[1], numpy.ndarray):
            raise cbor2.CBORDecodeValueError(f"Invalid raw data for multi-dimensional array tag ({tag.tag})")
        arr = tag.value[1].reshape(tag.value[0], order="C" if tag.tag == 40 else "F")
        return arr
    return None


# This variable is used to track the media types we are able to decode, mapping their names to
# decoding functions. The interface of the decoding function is to take two arguments: a bytes-like
# object containing the encoded data, and an encoding name (which may be None) given by the
# character set information (if any) included in the Content-Type header attached to the data.
AcceptTypes = {}

# This variable is used to track the codecs enabled on the server, mapping their names to
# decoding and encoding functions.  The interface for each should match that of json.loads() and
# json.dumps(), respectively.
Codecs = {}
Codec = namedtuple("Codec", ["decode", "encode"])


def decode_json(response_data, **kwargs):
    return json.loads(response_data, **kwargs)


def encode_json(obj, **kwargs):
    return json.dumps(obj, default=wrap_bytes_for_json, **kwargs)


Codecs["json"] = Codec(decode=decode_json, encode=encode_json)

if have_orjson:
    # If using orjson with NumPy, overload dumps with the right magic

    def decode_orjson(response_data, **kwargs):
        return orjson.loads(response_data, **kwargs)

    def encode_orjson(obj, **kwargs):
        if have_numpy:
            kwargs["option"] = kwargs.get("option", 0) | orjson.OPT_SERIALIZE_NUMPY
        return orjson.dumps(obj, default=wrap_bytes_for_json, **kwargs)

    Codecs["orjson"] = Codec(decode=decode_orjson, encode=encode_orjson)


def decode_json_client(response_data, encoding):
    if encoding is None:  # guess the typical default if unspecified
        encoding = "utf-8"

    def ohook(obj):
        if isinstance(obj, Mapping) and "bytes" in obj and (len(obj) == 1 or (len(obj) == 2 and "subtype" in obj)):
            try:
                return bytes(obj["bytes"])
            except ValueError as e:
                pass
        return TuberResult(obj)

    return decode_json(response_data.decode(encoding), object_hook=ohook)


AcceptTypes["application/json"] = decode_json_client


# Use cbor2 to handle CBOR, if available
if have_cbor:

    def decode_cbor(response_data, **kwargs):
        return cbor2.loads(response_data, tag_hook=cbor_tag_decode, **kwargs)

    def encode_cbor(obj, **kwargs):
        return cbor2.dumps(obj, default=cbor_augment_encode, **kwargs)

    Codecs["cbor"] = Codec(decode=decode_cbor, encode=encode_cbor)

    def decode_cbor_client(response_data, encoding):
        return decode_cbor(response_data, object_hook=lambda dec, data: TuberResult(data))

    AcceptTypes["application/cbor"] = decode_cbor_client



#!/usr/bin/env -S PYTHONPATH=.. pytest-3 -v

"""
Threading tests.

SQLAlchemy has a rational threading model, with helpers that easily provide a
session-per-thread policy. These tests are intended to ensure we stay on the
straight and narrow path.

In hidfmux, HardwareMap objects were thinly wrapped SQLAlchemy Session objects.
However, this broke the threading model - a single Session object couldn't be
passed to threaded code.

Luckily, SQLAlchemy provides [1] a scoped_session object that defers
construction of the underlying Session and keeps a cached copy in thread-local
storage.  This should give us a single Session object per thread.

[1]: https://docs.sqlalchemy.org/en/20/orm/contextual.html#sqlalchemy.orm.scoped_session
"""

import threading
import rfmux

def test_are_sessions_distinct_between_threads():
    '''Check that threads each get their own underlying Session / HardwareMap'''

    s = rfmux.load_session(
        """
        !HardwareMap
        - !CRS { serial: "0024" }
        """
    )

    results = []

    CHECKS = 10  # really only need 2 here

    def call(hwm):
        # Convert the HWM from a scoped_session proxy into an actual
        # session. Otherwise, every thread gets the same proxy and our
        # checks fail even if the underlying calls would have been proxied.
        hwm = hwm()

        # Now pass this back to the main thread so we can check identities
        results.append(hwm)

    threads = [ threading.Thread(target=call, args=(s, )) for n in range(CHECKS) ]

    for t in threads:
        t.start()

    for t in threads:
        t.join()

    assert len(results) == CHECKS
    for j, x in enumerate(results):
        for k, y in enumerate(results):
            assert (j==k) ^ (x is not y), "Identical Session/HardwareMap objects found in different Threads!"
